from enum import Enum
import random, time
from typing import Dict, Optional, Union
from feasytools import ReadOnlyTable, CDDiscrete, PDDiscrete, PDGamma, DTypeEnum
from ..locale import Lang
from ..traffic import EV, EVDict, RoadNet, DetectFiles
from .misc import VehicleType, random_diff, _TripInner, _EVInner, _xmlSaver
from .poly import PolygonMan

DictPDF = Dict[int, Union[PDDiscrete[int], None]]

TAZ_TYPE_LIST = ("Home", "Work", "Relax", "Other")

class RoutingCacheMode(Enum):
    """Routing cache mode"""
    NONE = 0  # No cache
    RUNTIME = 1 # Cache during runtime
    STATIC = 2 # Static cache in generation time

    def __str__(self):
        return ("None", "Runtime", "Static")[self.value]

    def __repr__(self):
        return self.value
    
class TripsGenMode(Enum):
    """Generation mode"""

    AUTO = "Auto"  # Automatic
    TYPE = "TYPE"  # Type-based
    POLY = "Poly"  # Polygon-based

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value

class EVsGenerator:    
    """Class to generate trips"""
    def __init__(self, CROOT: str, PNAME: str, seed,
            mode: TripsGenMode = TripsGenMode.AUTO,
            route_cache: RoutingCacheMode = RoutingCacheMode.NONE):
        """
        Initialization
            CROOT: Trip parameter folder
            PNAME: SUMO configuration folder
            seed: Random seed
        """
        _fn = DetectFiles(PNAME)
        random.seed(seed)
        self.vTypes = [VehicleType(**x) for x in ReadOnlyTable(CROOT + "/ev_types.csv",dtype=DTypeEnum.FLOAT32).to_list_of_dict()]
        # Define various functional area types
        self._route_cache_mode = route_cache
        self.net = RoadNet.load(_fn["net"])
        if mode == TripsGenMode.AUTO:
            if _fn.node_type: mode = TripsGenMode.TYPE
            elif _fn.poly and _fn.net and _fn.fcs: mode = TripsGenMode.POLY
            else: raise RuntimeError(Lang.ERROR_NO_TAZ_OR_POLY)
        if mode == TripsGenMode.TYPE:
            assert _fn.node_type, Lang.ERROR_NO_TAZ_OR_POLY
            self._mode = "type"
            self.dic_nodetype = {}
            with open(_fn.node_type, "r") as fp:
                for ln in fp.readlines():
                    name, lst = ln.split(":")
                    self.dic_nodetype[name.strip()] = [x.strip() for x in lst.split(",")]
        elif mode == TripsGenMode.POLY:
            assert _fn.poly and _fn.net and _fn.fcs, Lang.ERROR_NO_TAZ_OR_POLY
            self._mode = "poly"
            polys = PolygonMan(_fn.poly)
            self.dic_nodetype = {k:[] for k in TAZ_TYPE_LIST}
            for poly in polys:
                poly_type = poly.getConvertedType()
                if poly_type:
                    dist, node = self.net.find_nearest_node_with_distance(*poly.center())
                    # Ensure the edge is in the largest strongly connected component
                    if dist < 200 and self.net.is_node_in_largest_scc(node.id):
                        self.dic_nodetype[poly_type].append(node.id)
        else:
            raise RuntimeError(Lang.ERROR_NO_TAZ_OR_POLY)
        
        # Start time of first trip
        self.pdf_start_weekday = PDGamma(6.63, 65.76, 114.54)
        self.pdf_start_weekend = PDGamma(3.45, 84.37, 197.53)
        
        # Spatial transfer probability of weekday and weekend. 
        # key1 = from_type, key2 = time (0~95, each unit = 15min), value = CDF of (to_type1, to_type2, to_type3, to_type4)
        self.PSweekday:Dict[str, DictPDF] = {}
        self.PSweekend:Dict[str, DictPDF] = {}
        # Parking duration CDF of weekday and weekend.
        self.park_cdf_wd:Dict[str, CDDiscrete[int]] = {} 
        self.park_cdf_we:Dict[str, CDDiscrete[int]] = {}

        def read_trans_pdfs(path:str) -> DictPDF:
            tbwd = ReadOnlyTable(path, dtype=DTypeEnum.FLOAT32)
            times = [int(x) for x in tbwd.head[1:]]
            values = list(map(int, tbwd.col(0)))
            ret:DictPDF = {}
            for i in range(1, len(times)+1):
                weights = list(map(float, tbwd.col(i)))
                assert len(values) == len(weights)
                try:
                    ret[i] = PDDiscrete(values, weights)
                except ZeroDivisionError:
                    ret[i] = None
            return ret
        
        for dtype in TAZ_TYPE_LIST:
            self.PSweekday[dtype] = read_trans_pdfs(f"{CROOT}/space_transfer_probability/{dtype[0]}_spr_weekday.csv")
            self.PSweekend[dtype] = read_trans_pdfs(f"{CROOT}/space_transfer_probability/{dtype[0]}_spr_weekend.csv")
            self.park_cdf_wd[dtype] = CDDiscrete(f"{CROOT}/duration_of_parking/{dtype[0]}_spr_weekday.csv", True, int)
            self.park_cdf_we[dtype] = CDDiscrete(f"{CROOT}/duration_of_parking/{dtype[0]}_spr_weekend.csv", True, int)

        self.soc_pdf = PDDiscrete.fromCSVFileI(f"{CROOT}/soc_dist.csv", True)
    
    def __getPs(self, is_weekday: bool, dtype: str, time_index:int):
        return self.PSweekday[dtype].get(time_index, None) if is_weekday else self.PSweekend[dtype].get(time_index, None)
    
    def __getDest1(self, pfr: str, weekday: bool = True):
        """
        Get the destination of the trip secondary to the first trip
            pfr: Departure functional area type, such as "Home"
            weekday: Whether it is weekday or weekend
        Returns: 
            First trip: First departure time, arrival destination functional area type, such as "Work"
        """
        pdf = None
        while pdf is None:
            init_time = self.pdf_start_weekday.sample() if weekday else self.pdf_start_weekend.sample()
            if init_time >= 86400:
                continue
            # Time index (0~95, each unit = 15min)
            init_time_i = int(init_time / 15)
            pdf = self.__getPs(weekday, pfr, init_time_i)
        next_place = TAZ_TYPE_LIST[pdf.sample()]
        return int(init_time), next_place

    def __getDestA(self, from_type:str, init_time_i:int, weekday: bool):
        """
        Get the destination of the next trip for non-first trips
            from_type: Departure type, such as "Home"
            init_time_i: Time index (0~95, each unit = 15min)
            weekday: Whether it is weekday or weekend
        Returns:
            Destination type
        """
        cdf = self.__getPs(weekday, from_type, init_time_i)
        return "Home" if cdf is None else TAZ_TYPE_LIST[cdf.sample()]

    def __getNextNode(self, from_node:str, next_place_type:str) -> str:
        return random_diff(self.dic_nodetype[next_place_type], from_node)        
    
    def __genFirstTrip1(self, trip_id, weekday: bool = True):
        """
        Generate the first trip of the first day
            trip_id: Trip ID
            weekday: hether it is weekday or weekend
        Return a InnerTrip instance
        """
        from_Type = "Home"
        from_node = random.choice(self.dic_nodetype[from_Type])
        # Get departure time and destination area type
        depart_time_min, to_Type = self.__getDest1(from_Type, weekday)  # Minimum departure time is 0
        to_node = self.__getNextNode(from_node, to_Type)
        return _TripInner(trip_id, depart_time_min * 60, from_node, from_Type, to_node, to_Type)

    cdf_dict = {}

    def __genStopTimeIdx(self, from_type:str, weekday: bool):
        cdf = self.park_cdf_wd[from_type] if weekday else self.park_cdf_we[from_type]
        return int(cdf.sample() + 1)

    def __genTripA(
        self, trip_id:str, from_type:str, from_node:str, start_time:int, weekday: bool = True
    ) -> _TripInner:
        """
        Generate the second trip
            trip_id: Trip ID
            from_type: Departure area type, such as "Home"
            from_node: Departure roadside, such as "T6"
            start_time: Departure time of the first trip, in seconds since midnight
            weekday: Whether it is weekday or weekend
        """
        depart_time_min = 1440
        cnt = 0
        while depart_time_min >= 1440:  # If the departure time is after midnight, regenerate
            stop_time_idx = self.__genStopTimeIdx(from_type, weekday)
            depart_time_min = start_time // 60 + stop_time_idx * 15 + 20
            cnt += 1
            if cnt > 10:
                depart_time_min = start_time // 60 + 1
                break
        next_place_type = self.__getDestA(from_type, stop_time_idx, weekday)
        to_node = self.__getNextNode(from_node, next_place_type)
        return _TripInner(trip_id, depart_time_min * 60, from_node, from_type, to_node, next_place_type)

    def __genTripF(
        self, trip_id:str, from_type:str, from_node:str,
        start_time:int, first_node:str, weekday: bool = True,
    ):
        """
        Generate the third trip
            trip_id: Trip ID
            from_TAZ: Departure area TAZ type, such as "TAZ1"
            from_type: Departure area type, such as "Home"
            from_EDGE: Departure roadside, such as "gnE29"
            start_time: Departure time of the first trip, in seconds since midnight
            first_TAZ: First trip's destination area TAZ type, such as "TAZ2"
            first_EDGE: First trip's destination roadside, such as "gnE2"
            weekday: Whether it is weekday or weekend
        """
        if first_node == from_node:
            return None
        depart_time_min = 1440
        cnt = 0
        while depart_time_min >= 1440:  # If the departure time is after midnight, regenerate
            stop_time_idx = self.__genStopTimeIdx(from_type, weekday)
            depart_time_min = start_time // 60 + stop_time_idx * 15 + 20
            cnt += 1
            if cnt > 10:
                return None
        return _TripInner(trip_id, depart_time_min * 60, from_node, from_type, first_node, "Home")

    def __genTripsChain1(self, ev:_EVInner):  # vehicle_trip
        """
        Generate a full day of trips on the first day
            ev: vehicle instance
        """
        daynum = 0
        weekday = True
        trip_1 = self.__genFirstTrip1("trip0_1", weekday)
        trip_2 = self.__genTripA("trip0_2",trip_1.toT,trip_1.toN,trip_1.DPTT,weekday)
        trip_3 = self.__genTripF("trip0_3",trip_2.toT,trip_2.toN,trip_2.DPTT,trip_1.frN,weekday)
        
        ev._add_trip(daynum, trip_1)
        ev._add_trip(daynum, trip_2)
        if trip_3: # Trip3: if O==D, don't generate trip 3
            if trip_3.DPTT < 86400:  # If the departure time is after midnight, it is not valid
                ev._add_trip(daynum, trip_3)
            else:
                trip_2.toN = trip_1.frN

    def __genFirstTripA(self, trip_id:str, ev: _EVInner, weekday: bool = True):
        """
        Generate the first trip of a non-first day
            trip_id: Trip ID
            vehicle_node: Vehicle node, such as rootNode.getElementsByTagName("vehicle")[0]
            weekday: Whether it is weekday or weekend
        """
        trip_last = ev.trips[-1]
        from_node = trip_last.toN
        # Get departure time and destination area type
        from_Type = "Home"
        depart_time_min, to_Type = self.__getDest1(from_Type, weekday)
        to_node = self.__getNextNode(from_node, from_Type)
        return _TripInner(trip_id, depart_time_min * 60, from_node, from_Type, to_node, to_Type)

    def __genTripsChainA(self, ev: _EVInner, daynum: int = 1):  # vehicle_trip
        """
        Generate a full day of trips on a non-first day
        """
        weekday = (daynum - 1) % 7 + 1 in [1, 2, 3, 4, 5]
        trip2_1 = self.__genFirstTripA(f"trip{daynum}_1", ev, weekday)
        trip2_2 = self.__genTripA(f"trip{daynum}_2", trip2_1.toT,trip2_1.toN,trip2_1.DPTT,weekday)
        trip2_3 = self.__genTripF(f"trip{daynum}_3", trip2_2.toT,trip2_2.toN, trip2_2.DPTT,trip2_1.frN,weekday)
                    
        ev._add_trip(daynum, trip2_1)
        ev._add_trip(daynum, trip2_2)
        if trip2_3:
            if trip2_3.DPTT < 86400:  # If the departure time is after midnight, it is not valid
                ev._add_trip(daynum, trip2_3)
            else:
                trip2_2.toN = trip2_1.frN
        
    def __genEV(self, veh_id: str, day_count:int, **kwargs) -> _EVInner:
        '''
        Generate a full week of trips for a vehicle as an inner instance
        '''
        ev = _EVInner(veh_id, random.choice(self.vTypes), self.soc_pdf.sample()/100.0, **kwargs)
        self.__genTripsChain1(ev)
        for j in range(1, day_count + 1):
            self.__genTripsChainA(ev, j)
        return ev

    def genEV(self, veh_id: str, **kwargs) -> EV:
        """
        Generate a full week of trips for a vehicle.
        The generated vehicle is returned as an EV instance, and will not be held in the buffer.
            veh_id: ID of the vehicle
            v2g_prop: Proportion of users willing to participate in V2G
            omega: PDFunc | None = None,
            krel: PDFunc | None = None,
            ksc: PDFunc | None = None,
            kfc: PDFunc | None = None,
            kv2g: PDFunc | None = None
        """
        return self.__genEV(veh_id, **kwargs).toEV()

    def genEVs(
        self, N: int, fname: Optional[str] = None, day_count: int = 7, silent: bool = False, **kwargs
    ) -> EVDict:
        """
        Generate EV and trips of N vehicles.
        The generated vehicles are returned as an EVDict instance, and will be saved to the file if fname is provided.
        The vehicles will not be held in the buffer.
            N: Number of vehicles
            fname: Saved file name (if None, not saved)
            day_count: Number of days
            silent: Whether silent mode
            v2g_prop: Proportion of users willing to participate in V2G
            omega: PDFunc | None = None,
            krel: PDFunc | None = None,
            ksc: PDFunc | None = None,
            kfc: PDFunc | None = None,
            kv2g: PDFunc | None = None
        """
        st_time = time.time()
        last_print_time = 0
        saver = _xmlSaver(fname) if fname else None
        ret = EVDict()
        for i in range(0, N):
            ev = self.__genEV("v" + str(i), day_count,
                cache_route = self._route_cache_mode != RoutingCacheMode.NONE, **kwargs)
            ret.add(ev.toEV())
            if saver:
                saver.write(ev)
            if not silent and time.time()-last_print_time>1:
                print(f"\r{i+1}/{N}, {(i+1)/N*100:.2f}%", end="")
                last_print_time=time.time()
        if not silent:
            print(f"\r{N}/{N}, 100.00%")
            print(Lang.INFO_DONE_WITH_SECOND.format(round(time.time() - st_time, 1)))
        if saver:
            saver.close()
        return ret

class ManualEVsGenerator:
    """Class to manually add EVs to the buffer"""
    def __init__(self):
        self.__evs:Dict[str, _EVInner] = {}
    
    def addEV(self, vid:str, bcap_kWh:float, range_km:float, efc_rate_kW:float, 
            esc_rate_kW:float, max_V2G_kW:float, soc:float, omega:float, 
            krel:float, ksc:float, kfc:float, kv2g:float, cache_route:bool = False) -> _EVInner:
        """
        Add an EV to the generator's buffer
            vid: Vehicle ID
            vtype: Vehicle type
            soc: State of charge (0.0~1.0)
        """
        if vid in self.__evs:
            raise ValueError(f"Vehicle ID {vid} already exists.")
        ev = _EVInner(vid, VehicleType(id=-1, bcap_kWh=bcap_kWh, range_km=range_km,
            efc_rate_kW=efc_rate_kW, esc_rate_kW=esc_rate_kW, max_V2G_kW=max_V2G_kW), 
            soc, omega=omega, krel=krel, ksc=ksc, kfc=kfc, kv2g=kv2g, cache_route=cache_route)
        self.__evs[vid] = ev
        return ev
    
    def dumpEVs(self, fname: str):
        """
        Dump all EVs in the buffer to a file
            fname: File name
        """
        saver = _xmlSaver(fname)
        for ev in self.__evs.values():
            saver.write(ev)
        saver.close()
    
    def getEVs(self) -> EVDict:
        """
        Get all EVs in the buffer as an EVDict instance
        """
        ret = EVDict()
        for ev in self.__evs.values():
            ret.add(ev.toEV())
        return ret


class _locale:
    TRIPGEN_HELP_STR = "请阅读 https://hesl-seu.github.io/v2sim-wiki/#/zh_hans/v2sim/tools/gen_trip 上的使用说明。"
    CSGEN_HELP_STR = "请阅读 https://hesl-seu.github.io/v2sim-wiki/#/zh_hans/v2sim/tools/gen_cs 上的使用说明。"
    MAIN_HELP_STR = "请阅读 https://hesl-seu.github.io/v2sim-wiki/#/zh_hans/v2sim/tools/sim_single 上的使用说明。"
    PARA_HELP_STR = "请阅读 https://hesl-seu.github.io/v2sim-wiki/#/zh_hans/v2sim/tools/sim_para 上的使用说明。"

    ERROR_GENERAL = "错误: {}"
    ERROR_BAD_TYPE = "错误: 无效的数据类型{}."
    ERROR_ILLEGAL_CMD = "错误: 非法的命令行参数: {}."
    ERROR_CANNOT_USE_TOGETHER = "错误: 选项'{0}'和'{1}'不能同时使用."
    ERROR_UNKNOWN_CS_TYPE = "错误: 未知的充电站类型{}"
    ERROR_CMD_NOT_SPECIFIED = "错误: 必须指定参数'{}'."
    ERROR_SUMO_CONFIG_NOT_SPECIFIED = "错误: 未指定SUMO配置文件."
    ERROR_SUMO_N_VEH_NOT_SPECIFIED = "错误: 未指定车辆数."
    ERROR_FAIL_TO_OPEN = "错误: 无法打开文件{0}: {1}"
    ERROR_NET_FILE_NOT_SPECIFIED = "错误: 未指定路网文件."
    ERROR_TRIPS_FILE_NOT_FOUND = "错误: 未找到EV和行程文件."
    ERROR_FCS_FILE_NOT_FOUND = "错误: 未找到快充站文件."
    ERROR_SCS_FILE_NOT_FOUND = "错误: 未找到慢充站文件."
    ERROR_STATE_FILE_NOT_FOUND = "错误: 状态文件未找到, {0}."
    ERROR_ST_ED_TIME_NOT_SPECIFIED = "错误: 仿真起止时间未指定."
    ERROR_CLIENT_ID_NOT_SPECIFIED = "错误: 未指定客户端ID."
    ERROR_CONFIG_DIR_FILE_DUPLICATE = "错误: 配置文件夹中{0}有重复: {1}和{2}"
    ERROR_PLUGIN_INTERVAL = "错误: 未指定或非法的插件运行间隔"
    ERROR_STA_UNIDENTICAL_DATA_LEN_AND_HEAD = "错误: {0}的数据长度{1}与表头长度{2}不一致"
    ERROR_STA_REGISTERED = "错误: 统计项{0}已经注册过."
    ERROR_STA_ADDED = "错误: 统计项{0}已经添加过."
    ERROR_STA_LOG_ITEM = "错误: 记录项{0}时发生错误: {1}."
    ERROR_STA_CLOSE_ITEM = "错误: 关闭项{0}时发生错误: {1}."
    ERROR_STA_TIMELINE_NOT_FOUND = "错误: 未找到时间线文件."
    ERROR_FILE_EXISTS = "错误: 文件已存在: {}"
    ERROR_NUMBER_NOT_SPECIFIED = "错误: 当选择随机时必须指定n"
    ERROR_FILE_TYPE_NOT_SUPPORTED = "错误: 不支持的XML文件{}类型"
    ERROR_NO_TAZ_OR_POLY = "错误: 未找到TAZ或POLY文件. 如果要使用POLY模式, 还必须指定FCS文件."
    ERROR_RANDOM_CANNOT_EXCLUDE = "错误: 序列抽样中的所有元素都被排除"
    ERROR_ROUTE_NOT_FOUND = "错误: 无法找到从{0}到{1}的路径"
    ERROR_INVALID_CACHE_ROUTE = "错误: 无效的寻路缓存模式: {0}"
    ERROR_INVALID_TRIP_GEN_MODE = "错误: 无效的行程生成模式: {0}"
    ERROR_CS_NODE_NOT_EXIST = "错误: 充电站所在节点{0}不存在"

    WARN_EXT_LOAD_FAILED = "警告: {0}是Python文件, 但无法作为包加载: {1}"
    WARN_EXT_INVALID_PLUGIN = "警告: {0}的plugin_exports无效, 无法作为插件导入: {1}"
    WARN_EXT_INVALID_STA = "警告: {0}的sta_exports无效, 无法作为统计项导入: {1}"
    WARN_MAIN_SHOW_MEANINGLESS = "警告: show选项在Windows下没有意义, 请在v2sim/traffic/win_vis.py中调整WINDOWS_VISUALIZE来改变可见性级别."
    WARN_MAIN_DEBUG_MEANINGLESS = "警告: debug选项在命令行模式下没有意义, 将自动关闭."
    WARN_MAIN_GUI_NOT_FOUND = "警告: 未找到GUI模块, 请检查是否安装了tkinter库. 将切换到命令行模式."
    WARN_SIM_COMM_FAILED = "警告: 与主进程通信失败."
    WARN_CS_NOT_IN_SCC = "警告: 快充站{}不在最大的强连通分量中."
    WARN_SCC_TOO_SMALL = "警告: 最大的强连通分量太小, 只有{0}/{1}条边在内."
    WARN_SIMULATION_STALLED = "警告: 仿真可能出现问题: 连续10步的平均速度<0.001 m/s. 结果中的行程和充电负荷可能表现为截断."

    INFO_DONE_WITH_DURATION = "已完成. 用时: {}."
    INFO_DONE_WITH_SECOND = "已完成. 用时{:.1f}秒."
    INFO_SUMO = "  SUMO: {}"
    INFO_NET = "  路网: {}"
    INFO_TRIPS = "  行程: {0}, {1}辆EV"
    INFO_FCS = "  快充: {0}, {1}个站点"
    INFO_SCS = "  慢充: {0}, {1}个站点"
    INFO_TIME = "  起止时间: {0} ~ {1} (中断于{2}), 步长：{3}"
    INFO_PLG = "  插件: {0} - {1}"
    INFO_REGEN_SCS = "慢充站已重新生成."
    INFO_REGEN_FCS = "快充站已重新生成."
    INFO_REGEN_VEH = "EV和行程已重新生成."

    CORE_NO_RUN = "这是仿真系统的核心模块。不要直接运行此文件。请改用sim_single.py或sim_para.py。"

    MAIN_LS_TITLE_PLG = "=== 插件 ==="
    MAIN_LS_TITLE_STA = "=== 统计项 ==="
    MAIN_SIM_START = "仿真开始, 按Ctrl-C中断"
    MAIN_SIGINT = "收到Ctrl-C退出信号, 提前退出"
    MAIN_SIM_DONE = "仿真结束. 用时: {}"
    MAIN_SIM_PROG = "进度: {0:.2f}%, {1}/{2}. 已用时: {3}, 预计剩余时间: {4}"

    PARA_SIM_SKIP_LIST = "跳过: {}"
    PARA_SIM_DONE_PARA = "并行部分完成. 用时: {}"
    PARA_SIM_START_SERIAL = "执行非并行任务..."
    PARA_SIM_DONE_SERIAL = "串行部分完成. 用时: {}"
    PARA_SIM_PROG = "进度: {0:.2f}%, {1}已用时: {2}, 预计剩余时间: {3}"

    PLOT_GRP_EMPTY_SEC_LIST = "秒列表为空"
    PLOT_GRP_START_TIME_EXCEED = "MinAvgGrouper的起始时间({0})超过了数据记录的时间{1}"
    PLOT_GRP_X_LABEL = "第{0}天 {1:02}:{2:02}"
    PLOT_GRP_DATA_UNMATCH = "时间轴长度{}和数据数量{}不匹配"

    CSLIST_INVALID_ELEMENT = "csList中的元素必须是FCS或SCS"
    CSLIST_MIXED_ELEMENT = "CSList禁止同时存在FCS和SCS元素, 将ALLOW_MIXED_CSTYPE_IN_CSLIST设为True以移除此限制"
    CSLIST_INVALID_TAG = "用xml文件初始化CSList时, 遇到无效的tag{}"
    CSLIST_PBUY_NOT_SPECIFIED = "用xml文件初始化CSList时, 未指定购电价格"
    CSLIST_INVALID_INIT_PARAM = "无效的CSList初始化参数类型"
    CSLIST_KDTREE_DISABLED = "    由于存在无效充电站位置，KD树已禁用，将不能使用select_near功能"

    CPROC_ARRIVE = "到达"
    CPROC_ARRIVE_CS = "快充开始"
    CPROC_DEPART = "出发"
    CPROC_DEPART_DELAY = "延误"
    CPROC_DEPART_CS = "快充完成"
    CPROC_DEPART_FAILED = "故障"
    CPROC_FAULT_DEPLETE = "耗尽"
    CPROC_FAULT_NOCHARGE = "无可用快充站"
    CPROC_FAULT_REDIRECT = "重定向"
    CPROC_WARN_SMALLCAP = "警告"
    CPROC_JOIN_SCS = "加入慢充站"
    CPROC_LEAVE_SCS = "离开慢充站"

    CPROC_INFO_ARRIVE = "车辆{0}已到达{1}, {2}, 下一行程: {3}"
    CPROC_INFO_ARRIVE_0 = "不充电"
    CPROC_INFO_ARRIVE_1 = "开始慢充"
    CPROC_INFO_ARRIVE_2 = "无可用慢充桩, 充电失败"
    CPROC_INFO_ARRIVE_CS = "车辆{0}到达{1}, 排队充电."
    CPROC_INFO_DEPART = "车辆{0}出发, 行程{1}."
    CPROC_INFO_DEPART_WITH_DELAY = " 延迟{0}秒."
    CPROC_INFO_DEPART_WITH_CS = " 将在{0}充电, 参数 = {1}."
    CPROC_INFO_DEPART_DELAY = "车辆{0}未能出发, 因为现有电量{1}, 而需要电量{2}. 延迟{3}秒再试."
    CPROC_INFO_DEPART_CS = "车辆{0}在{1}充完电, 继续前往{2}."
    CPROC_INFO_DEPART_FAILED = "车辆{0}由于电量不足无法发车. 需要{1}但只有{2}. 将会在{4}秒后传送到{3}."
    CPROC_INFO_FAULT_DEPLETE = "车辆{0}电量耗尽. 将会在{2}秒后传送到{1}."
    CPROC_INFO_FAULT_NOCHARGE = "车辆{0}没有电量. 将会传送到{1}."
    CPROC_INFO_FAULT_REDIRECT = "车辆{0}在{1}无法充电, 原因是电量不足. 正在重定向到{2}."
    CPROC_INFO_WARN_SMALLCAP = "车辆{0}的电池容量为{1}, 少于完成行程的需求{2}. 将在行程中耗尽电量."
    CPROC_INFO_JOIN_SCS = "车辆{0}加入慢充站{1}."
    CPROC_INFO_LEAVE_SCS = "车辆{0}离开慢充站{1}."

    PLG_REGISTERED = "插件{}已注册."
    PLG_DEPS_MUST_BE_STRLIST = "插件依赖必须是字符串列表."
    PLG_NOT_SUBCLASS = "插件{}不是PluginBase的子类."
    PLG_DEPS_NOT_REGISTERED = "插件{0}依赖的插件{1}尚未注册."
    PLG_INTERVAL_NOT_SPECIFIED = "插件{0}未指定或非法的运行间隔."
    PLG_NOT_EXIST = "文件{}不存在. 跳过插件加载."
    PLG_NOT_EXIST_OR_BAD = "文件{}不存在或无法导入."
    PLG_INVALID_PLUGIN = "无效插件{}"
    PLG_DEPS_NOT_LOADED = "插件{0}依赖的插件{1}尚未加载."
    PLG_ALREADY_EXISTS = "插件{}已经存在于插件列表."

    PLOT_FONT = "SimHei"
    PLOT_FONT_SIZE_SMALL = "13"
    PLOT_FONT_SIZE_MEDIUM = "15"
    PLOT_FONT_SIZE_LARGE = "18"
    PLOT_STR_ALL = "全部"
    PLOT_STR_FAST = "快"
    PLOT_STR_SLOW = "慢"
    PLOT_NOT_SUPPORTED = "既有数据不支持绘制{}的图形"

    ADV_PLOT_TITLE = "高级绘图工具 - 输入'help'获取帮助"
    ADV_PLOT_HELP = '''命令:
    plot <series_name> [<label> <color> <linestyle> <side>]: 绘制一个数据序列
    title <title>: 设置标题
    xlabel <label>: 设置x轴标签
    yleftlabel/ylabel <label>: 设置左y轴标签
    yrightlabel <label>: 设置右y轴标签
    yticks <ticks> [<labels>]: 设置y轴刻度
    legend <loc>: 显示图例
    save <path>: 保存图像到指定路径
    exit: 退出绘图程序
示例:
    plot "results:cs_load:CS1" "CS1 Load" "blue" "-" "left"
    plot "results:cs_load:CS2" "CS2 Load" "red" "--" "left"
    title "CS1 & CS2 Load"
    xlabel "Time"
    yleftlabel "Load/kWh"
    legend
    save "test.png"
序列名称的格式为"<results>:<attribute>:<instances>:<start_time>", 其中
    "results"是结果文件夹名,
    "attribute"是要绘制的属性,
    "instances"是实例名称，例如
        充电站(CS): "CS1", "CS2", "<all>", "<fast>", "<slow>"
        电动汽车(EV): "v1", "v2"
        发电机: "G1", "G2"
        母线: "B1", "B2"
        等,
    "start_time"是绘图的起始时间, 默认为0, 该项可以省略
你可以从文件中读取命令。通过将文件名作为命令行参数传递给程序来实现。
'''

    PLOT_EV = "电动汽车: {0}"
    PLOT_FCS_ACC_TITLE = "快充站: 总负荷"
    PLOT_SCS_ACC_TITLE = "慢充站: 总负荷"
    PLOT_YLABEL_POWERKW = "功率(kW)"
    PLOT_YLABEL_POWERMW = "功率(MW或Mvar)"
    PLOT_YLABEL_VOLTKV = "电压(kV)"
    PLOT_YLABEL_CURRENT = "电流(kA)"
    PLOT_YLABEL_COST = "钱($)"
    PLOT_YLABEL_COUNT = "数量"
    PLOT_YLABEL_SOC = "荷电状态(%)"
    PLOT_YLABEL_CURTAIL = "舍弃率(%)"
    PLOT_YLABEL_STATUS = "车辆状态"
    PLOT_YLABEL_PRICE = "价格($/kWh)"
    PLOT_XLABEL_TIME = "时间"
    PLOT_FCS_TITLE = "快充站: {0}"
    PLOT_SCS_TITLE = "慢充站: {0}"
    PLOT_BUS_TOTAL = "总母线负荷"
    PLOT_GEN_TOTAL = "总发电机"
    PLOT_LINE = "线路: {0}"
    PLOT_GEN = "发电机: {0}"
    PLOT_BUS = "母线: {0}"
    PLOT_PVW = "光伏或风机: {0}"
    PLOT_ESS = "储能系统: {0}"

    PLOT_TOOL_CLEARING = "正在清除{0}..."
    

    CONVERT_ERROR_MISSING_PATHS = "错误: 请使用'-i'和'-o'参数指定输入输出文件夹。"
    CSQUERY_KEY_REQUIRED = "请在命令行中使用'--key'提供高德地图密钥"
    BAD_TRIP_OD = "行程起始节点必须与前一个行程的结束节点相同，但车辆{2}的行程{3}的起点{0}不是{1}"
    BAD_TRIP_DEPART_TIME = "行程出发时间必须按升序排列，但车辆{2}的行程{3}的出发时间{0}早于{1}"
    PARA_WORLDS = "已创建{0}个并行仿真."
    SINGLE_WORLD = "已创建单个串行仿真."
    ROUTE_ALGO_NOT_SUPPORTED = "寻路算法'{0}'无效. 可用项为: dijkstra, astar"
    VEH_NOT_FOUND = "车辆{0}不在仿真中."
    VEH_HAS_NO_LINK = "车辆{0}不在任何道路上."
    NO_AVAILABLE_FCS = "目前没有可用快充站，请调整配置."
    EV_LOAD_ERROR = "无法加载车辆文件{0}."
    NODE_EXISTS = "节点{0}已存在."
    NODE_NOT_FOUND = "节点{0}不存在."
    EDGE_EXISTS = "边{0}已存在."
    EDGE_NOT_FOUND = "边{0}不存在."
    INVALID_SUMO_NETWORK = "无效的SUMO网络: {0}"
    UNKNOWN_NET_FORMAT = "未知的网络格式: {0}. 可用项: raw, sumo, auto."
    GIL_NOT_DISABLED = "警告: 并行仿真需要在禁用GIL的情况下启动Python. 已回退到串行仿真."
    NO_GEO_PROJ = "网络不提供地理投影"
    SIMULATION_MAY_STALL = "仿真可能停滞：时间 {0} 处平均速度 < 0.001 米/秒。"
    PY_VERSION_MISMATCH_TI = "TrafficInst 的 Python 版本不匹配：预期 {0}，实际 {1}。"
    PY_VERSION_MISMATCH_PLG = "插件状态的 Python 版本不兼容：保存时版本 {0}，当前版本 {1}。"
    PICKLER_MISMATCH_TI = "TrafficInst 的 Pickler 不匹配：预期 {0}，实际 {1}。"
    PICKLER_MISMATCH_PLG = "插件状态的 Pickler 不兼容：保存时 {0}，当前 {1}。"
    SAVED_STATE_COPIED = "保存的状态文件已复制到项目文件夹。"
    CASE_FILE_COPIED = "案例文件已复制到结果文件夹。"
    NO_HOST_EXISTS = "未在多进程模式下工作。主机不存在。"
    ALT_COMMAND_NOT_SUPPORTED = "当指定了initial_state时，无法使用alt_command。"
    INVALID_PLUGIN_STATES = "保存的状态中存在无效的插件状态：{0}。"


import re
from importlib.metadata import distribution, PackageNotFoundError
from typing import List, Tuple, Optional

def _parse_requirement(line: str) -> Optional[Tuple[str, Optional[str]]]:
    line = line.strip()
    if not line or line.startswith(('#')):
        return None
    
    match = re.match(r'^([a-zA-Z0-9\-_\.]+)([=<>!~]=?.*)?$', line)
    if not match:
        return None
    
    package = match.group(1).lower()
    version_spec = match.group(2).strip() if match.group(2) else None
    return package, version_spec

def _version_satisfies(installed_version: str, version_spec: str) -> bool:
    from packaging import version
    from packaging.specifiers import SpecifierSet

    try:
        spec = SpecifierSet(version_spec, prereleases=True)
        return spec.contains(version.parse(installed_version))
    except:
        return False

def _check_requirements(requirements_file: str) -> List[Tuple[str, str, str]]:
    violations = []
    
    try:
        with open(requirements_file, 'r') as f:
            requirements = [_parse_requirement(line) for line in f]
            requirements = [r for r in requirements if r is not None]
    except FileNotFoundError:
        print(f"Error: {requirements_file} not found")
        return violations
    
    for package, version_spec in requirements:
        try:
            dist = distribution(package)
            installed_version = dist.version
            
            if version_spec and not _version_satisfies(installed_version, version_spec):
                violations.append((package, version_spec, installed_version))
                
        except PackageNotFoundError:
            violations.append((package, version_spec, "NOT INSTALLED"))
    
    return violations

def check_requirements(file = 'requirements.txt'):
    violations = _check_requirements(file)
    
    if not violations: return
    print("Some packages do not meet the requirements:")
    
    max_name_len = max(len(v[0]) for v in violations)
    max_spec_len = max(len(v[1] or 'any') for v in violations)
    
    for package, spec, installed in violations:
        if installed == "NOT INSTALLED":
            print(f"  {package:<{max_name_len}} - Required: {spec or 'any':<{max_spec_len}} | Current: Not installed")
        else:
            print(f"  {package:<{max_name_len}} - Required: {spec or 'any':<{max_spec_len}} | Current: {installed}")
    
    print("Please install them via 'pip install -r requirements.txt'")
    exit(1)

def check_requirements_gui(file = 'requirements.txt'):
    violations = _check_requirements(file)
    
    if not violations: return

    import tkinter as tk
    from tkinter import scrolledtext
    import subprocess
    import sys, os
    import threading
    
    root = tk.Tk()
    root.title("Requirements Checker")

    msg = "Some packages do not meet the requirements:\n\n"
    max_name_len = max(len(v[0]) for v in violations)
    max_spec_len = max(len(v[1] or 'any') for v in violations)

    for package, spec, installed in violations:
        if installed == "NOT INSTALLED":
            msg += f"{package:<{max_name_len}} - Required: {spec or 'any':<{max_spec_len}} | Current: Not installed\n"
        else:
            msg += f"{package:<{max_name_len}} - Required: {spec or 'any':<{max_spec_len}} | Current: {installed}\n"

    command = f"{sys.executable} -m pip install -r {file}"
    msg += f"\nPlease install them via '{command}'"

    text = scrolledtext.ScrolledText(root, width=80, height=15)
    text.insert(tk.END, msg)
    text.config(state=tk.DISABLED)
    text.pack(padx=10, pady=10)

    btn_frame = tk.Frame(root)
    btn_frame.pack(pady=(0, 10))

    btn = tk.Button(btn_frame, text="Exit", command=root.destroy)
    btn.pack(side=tk.LEFT, padx=5)

    close = True
    def run_command():
        for widget in btn_frame.winfo_children():
            widget.pack_forget()
        installing_label = tk.Label(btn_frame, text="Installing...", fg="blue")
        installing_label.pack(side=tk.LEFT, padx=5)
        root.update()

        def install_and_restart():
            nonlocal close
            close = False
            subprocess.call(command, shell=True)
            root.destroy()
            os.execl(sys.executable, sys.executable, *sys.argv)

        threading.Thread(target=install_and_restart, daemon=True).start()

    install_btn = tk.Button(btn_frame, text="Install and Retry", command=run_command)
    install_btn.pack(side=tk.LEFT, padx=5)

    root.mainloop()
    if close: exit(1)

if __name__ == "__main__":
    check_requirements()
    print("All packages meet the requirments.")


import copy
import importlib
from feasytools import LangConfig


LangConfig.SetAppName("v2simux")


class Lang:    
    TRIPGEN_HELP_STR = "Read https://hesl-seu.github.io/v2sim-wiki/#/v2sim/tools/gen_trip to see the usage."
    CSGEN_HELP_STR = "Read https://hesl-seu.github.io/v2sim-wiki/#/v2sim/tools/gen_cs to see the usage."
    MAIN_HELP_STR = "Read https://hesl-seu.github.io/v2sim-wiki/#/v2sim/tools/sim_single to see the usage."
    PARA_HELP_STR = "Read https://hesl-seu.github.io/v2sim-wiki/#/v2sim/tools/sim_para to see the usage."
    
    ERROR_GENERAL = "Error: {}"
    ERROR_BAD_TYPE = "Error: Invalid data type '{}'."
    ERROR_NO_SUPPORTED_LANG = "Error: No supported language found."
    ERROR_UNSUPPORTED_LANG = "Error: Unsupported language '{}'."
    ERROR_ILLEGAL_CMD = "Error: Illegal command line parameter :{}."
    ERROR_CANNOT_USE_TOGETHER = "Error: Options '{0}' and '{1}' cannot be used together."
    ERROR_UNKNOWN_CS_TYPE = "Error: Unknown charging station (CS) type '{}'"
    ERROR_CMD_NOT_SPECIFIED = "Error: Option '{}' must be specified."
    ERROR_SUMO_CONFIG_NOT_SPECIFIED = "Error: SUMO configuration file not specified."
    ERROR_SUMO_N_VEH_NOT_SPECIFIED = "Error: Number of vehicles not specified."
    ERROR_FAIL_TO_OPEN = "Error: Failed to open file {0}: {1}"
    ERROR_NET_FILE_NOT_SPECIFIED = "Error: Road network file not specified."
    ERROR_TRIPS_FILE_NOT_FOUND = "Error: EV & Trips file not found."
    ERROR_FCS_FILE_NOT_FOUND = "Error: Fast charging station file not found."
    ERROR_SCS_FILE_NOT_FOUND = "Error: Slow charging station file not found."
    ERROR_STATE_FILE_NOT_FOUND = "Error: State file not found, {0}"
    ERROR_ST_ED_TIME_NOT_SPECIFIED = "Error: Start and end time not specified."
    ERROR_CLIENT_ID_NOT_SPECIFIED = "Error: Client ID not specified."
    ERROR_CONFIG_DIR_FILE_DUPLICATE = "Error: Duplicate item '{0}' in configuration directory: {1} and {2}"
    ERROR_PLUGIN_INTERVAL = "Error: Plugin interval not specified or invalid."
    ERROR_STA_UNIDENTICAL_DATA_LEN_AND_HEAD = "Error: Data length {1} of {0} is not identical to header length {2}"
    ERROR_STA_REGISTERED = "Error: Statistic item {0} is already registered."
    ERROR_STA_ADDED = "Error: Statistic item {0} is already added."
    ERROR_STA_LOG_ITEM = "Error: An error occurred when logging item {0}: {1}."
    ERROR_STA_CLOSE_ITEM = "Error: An error occurred when closing item {0}: {1}."
    ERROR_STA_TIMELINE_NOT_FOUND = "Error: Time line file not found."
    ERROR_FILE_EXISTS = "Error: File already exists: {}"
    ERROR_NUMBER_NOT_SPECIFIED = "Error: n must be specified when selecting randomly"
    ERROR_FILE_TYPE_NOT_SUPPORTED = "Error: XML File {} type not supported."
    ERROR_NO_TAZ_OR_POLY = "Error: No TAZ or polygon file found. If you want use poly mode, you have to specify a fcs file in addition."
    ERROR_RANDOM_CANNOT_EXCLUDE = "Error: No element in the sequence is different from the excluded value"
    ERROR_ROUTE_NOT_FOUND = "Error: Unable to find a path from {0} to {1}"
    ERROR_INVALID_CACHE_ROUTE = "Error: Invalid cache route: {0}"
    ERROR_INVALID_TRIP_GEN_MODE = "Error: Invalid trip generation mode: {0}"
    ERROR_CS_NODE_NOT_EXIST = "Error: CS node {0} does not exist in the road network."

    WARN_EXT_LOAD_FAILED = "Warning: {0} is a Python file, but cannot be loaded as a package: {1}"
    WARN_EXT_INVALID_PLUGIN = "Warning: {0}'s plugin_exports is invalid, cannot be imported as a plugin: {1}"
    WARN_EXT_INVALID_STA = "Warning: {0}'s sta_exports is invalid, cannot be imported as a statistic item: {1}"
    WARN_MAIN_SHOW_MEANINGLESS = "Warning: 'show' option is meaningless in Windows, please adjust WINDOWS_VISUALIZE in v2sim/traffic/win_vis.py to change visibility level"
    WARN_MAIN_DEBUG_MEANINGLESS = "Warning: 'debug' option is meaningless in command line mode, will be turned off automatically."
    WARN_MAIN_GUI_NOT_FOUND = "Warning: GUI module not found, please check if tkinter library is installed. Will switch to command line mode."
    WARN_SIM_COMM_FAILED = "Warning: Failed to communicate with simulation process."
    WARN_CS_NOT_IN_SCC = "Warning: some CS {} are not in the largest SCC"
    WARN_SCC_TOO_SMALL = "Warning: the largest SCC is too small, only {0} of {1} edges are included"
    WARN_SIMULATION_STALLED = "Warning: Simulation may stall: average speed < 0.001 m/s for 10 consecutive steps. Trips and charging processes may be eliminated in the results."

    INFO_DONE_WITH_DURATION = "Done. Duration: {}."
    INFO_DONE_WITH_SECOND = "Done. Duration: {:.1f} second(s)."
    INFO_SUMO = "  SUMO: {}"
    INFO_NET = "  Road Network: {}"
    INFO_TRIPS = "  EV & Trips: {0}, Number = {1}"
    INFO_FCS = "  Fast Charging Station: {0}, Number = {1}"
    INFO_SCS = "  Slow Charging Station: {0}, Number = {1}"
    INFO_TIME = "  Start & End Time: {0} ~ {1} (Break at {2}), Step Length = {3}"
    INFO_PLG = "  Plugin: {0} - {1}"
    INFO_REGEN_SCS = "Slow charging stations regenerated."
    INFO_REGEN_FCS = "Fast charging stations regenerated."
    INFO_REGEN_VEH = "EVs & trips regenerated."

    CORE_NO_RUN = "This is the core module of the simulation system. Do not run this file directly. Use sim_single.py or sim_para.py instead."

    MAIN_LS_TITLE_PLG = "=== Plugins ==="
    MAIN_LS_TITLE_STA = "=== Statistic Items ==="
    MAIN_SIM_START = "Simulation started. Press Ctrl-C to stop."
    MAIN_SIGINT = "Received Ctrl-C signal, exiting prematurely."
    MAIN_SIM_DONE = "Simulation done. Duration: {}"
    MAIN_SIM_PROG = "Progress: {0:.2f}%, {1}/{2}. Elapsed: {3}, ETA: {4}"

    PARA_SIM_SKIP_LIST = "Skip list: {}"
    PARA_SIM_DONE_PARA = "Parallel part done. Duration: {}"
    PARA_SIM_START_SERIAL = "Execute non-parallel tasks..."
    PARA_SIM_DONE_SERIAL = "Serial part done. Duration: {}"
    PARA_SIM_PROG = "Progress: {0:.2f}%, {1}Elapsed: {2}, ETA: {3}"

    PLOT_GRP_EMPTY_SEC_LIST = "Empty second list."
    PLOT_GRP_START_TIME_EXCEED = "Start time {0} of MinAvgGrouper exceeds the last second {1} in the second list."
    PLOT_GRP_X_LABEL = "Day {0} {1:02}:{2:02}"
    PLOT_GRP_DATA_UNMATCH = "Time line length {} does not match data length {}."

    CSLIST_INVALID_ELEMENT = "Elements in csList must be FCS or SCS"
    CSLIST_MIXED_ELEMENT = "Elements in csList must be all FCS or all SCS. Set 'ALLOW_MIXED_CSTYPE_IN_CSLIST' to 'True' to remove this restriction."
    CSLIST_INVALID_TAG = "Invalid tag {} when initializing CSList with xml file."
    CSLIST_PBUY_NOT_SPECIFIED = "Purchase price not specified when initializing CSList with xml file."
    CSLIST_INVALID_INIT_PARAM = "Invalid initialization parameter type for CSList."
    CSLIST_KDTREE_DISABLED = "    KDTree is disabled due to invalid CS position. Cannot find nearest CS."

    CPROC_ARRIVE = "Arrival"
    CPROC_ARRIVE_CS = "Fast Charging Start"
    CPROC_DEPART = "Departure"
    CPROC_DEPART_DELAY = "Departure Delayed"
    CPROC_DEPART_CS = "Fast Charging Done"
    CPROC_DEPART_FAILED = "Departure Failed"
    CPROC_FAULT_DEPLETE = "Depletion"
    CPROC_FAULT_NOCHARGE = "No FCS available"
    CPROC_FAULT_REDIRECT = "Redirect"
    CPROC_WARN_SMALLCAP = "Warning"
    CPROC_JOIN_SCS = "Join Slow Charging Station"
    CPROC_LEAVE_SCS = "Leave Slow Charging Station"

    CPROC_INFO_ARRIVE = "Vehicle {0} arrived at {1}. {2}. Next trip: {3}"
    CPROC_INFO_ARRIVE_0 = "No charging"
    CPROC_INFO_ARRIVE_1 = "Start slow charging"
    CPROC_INFO_ARRIVE_2 = "No available slow charging station, charging failed"
    CPROC_INFO_ARRIVE_CS = "Vehicle {0} arrive at {1}. Queue to charge."
    CPROC_INFO_DEPART = "Vehicle {0} depart from {1}."
    CPROC_INFO_DEPART_WITH_DELAY = " With a delay of {0} second(s)."
    CPROC_INFO_DEPART_WITH_CS = " Will charge at {0}, params = {1}."
    CPROC_INFO_DEPART_DELAY = "Vehicle {0} currently has a battery level of {1}, failing the requirements of battery level {2}. Wait {3} seocnd(s) to retry."
    CPROC_INFO_DEPART_CS = "Vehicle {0} finished charging at {1}, continuing the trip to {2}."
    CPROC_INFO_DEPART_FAILED = "Vehicle {0} failed to depart due to insufficient battery level. Required: {1}, Current: {2}. Will be teleported to {3} after {4} second(s)."
    CPROC_INFO_FAULT_DEPLETE = "Vehicle {0} depleted. Will be teleported to {1} after {2} seconds."
    CPROC_INFO_FAULT_NOCHARGE = "Vehicle {0} has no charge. Will be teleported to {1}."
    CPROC_INFO_FAULT_REDIRECT = "Vehicle {0} cannot charge at {1} due to insufficient battery level. Redirecting to {2}."
    CPROC_INFO_WARN_SMALLCAP = "Vehicle {0} has a battery capacity of {1}, which is less than the required {2}. Will cause a depletion during the trip."
    CPROC_INFO_JOIN_SCS = "Vehicle {0} joined slow charging station at {1}."
    CPROC_INFO_LEAVE_SCS = "Vehicle {0} left slow charging station at {1}."

    PLG_REGISTERED = "Plugin '{}' has already been registered."
    PLG_DEPS_MUST_BE_STRLIST = "Dependencies must be a list of strings."
    PLG_NOT_SUBCLASS = "Plugin '{}' is not a subclass of PluginBase."
    PLG_DEPS_NOT_REGISTERED = "Plugin '{0}' depends on plugin '{1}', which has not been registered."
    PLG_INTERVAL_NOT_SPECIFIED = "Plugin {}'s interval is not specified."
    PLG_NOT_EXIST = "File {} does not exist. Skip loading plugins."
    PLG_NOT_EXIST_OR_BAD = "File {} does not exist or has been corrupted."
    PLG_INVALID_PLUGIN = "Invalid plugin '{}'."
    PLG_DEPS_NOT_LOADED = "Plugin '{0}' depends on plugin '{1}', which has not been loaded."
    PLG_ALREADY_EXISTS = "Plugin '{}' already exists."
    
    PLOT_FONT = "Arial"
    PLOT_FONT_SIZE_SMALL = "12"
    PLOT_FONT_SIZE_MEDIUM = "14"
    PLOT_FONT_SIZE_LARGE = "18"
    PLOT_STR_ALL = "All"
    PLOT_STR_FAST = "Fast"
    PLOT_STR_SLOW = "Slow"
    PLOT_NOT_SUPPORTED = "The existing data does not support drawing the graph of {}"

    ADV_PLOT_TITLE = "Advanced Plotting Tool - Type 'help' for help"
    ADV_PLOT_HELP = '''Commands:
    plot <series_name> [<label> <color> <linestyle> <side>]: Add a series to the plot
    title <title>: Set the title of the plot
    xlabel <label>: Set the x-axis label
    yleftlabel/ylabel <label>: Set the left y-axis label
    yrightlabel <label>: Set the right y-axis label
    yticks <ticks> [<labels>]: Set the y-axis ticks and labels
    legend <loc>: Set the legend location
    save <path>: Save the plot to the path
    exit: Exit the configuration
Example:
    plot "results:cs_load:CS1" "CS1 Load" "blue" "-" "left"
    plot "results:cs_load:CS2" "CS2 Load" "red" "--" "left"
    title "CS1 & CS2 Load"
    xlabel "Time"
    yleftlabel "Load/kWh"
    legend
    save "test.png"
The series name should be in the format of "<results>:<attribute>:<instances>:<starting_time>", where
    "results" is the result folder, 
    "attribute" is the attribute to be plotted,
    "instances" is the instance name, such as 
        "CS1", "CS2", "<all>", "<fast>", "<slow>" for CS,
        "v1", "v2" for EV,
        "G1", "G2" for generator,
        "B1", "B2" for bus, 
        etc.,
    "starting_time" is the starting time of the series. "0" by default. This is optional.
You can load the commands from a file as an argument in the command prompt/terminal.
'''
    ADV_PLOT_NO_SERIES = "Series not provided"
    ADV_PLOT_NO_XLABEL = "X-axis label not provided"
    ADV_PLOT_NO_YLABEL = "Y-axis label not provided"
    ADV_PLOT_NO_TITLE = "Title not provided"
    ADV_PLOT_BAD_COMMAND = "Bad command: {}"

    PLOT_EV = "Electric Vehicle: {0}"
    PLOT_FCS_ACC_TITLE = "Fast Charging Station: Total Load"
    PLOT_SCS_ACC_TITLE = "Slow Charging Station: Total Load"
    PLOT_YLABEL_POWERKW = "Power (kW)"
    PLOT_YLABEL_POWERMW = "Power (MW or Mvar)"
    PLOT_YLABEL_COST = "Money ($)"
    PLOT_YLABEL_VOLTKV = "Voltage (kV)"
    PLOT_YLABEL_CURRENT = "Current (kA)"
    PLOT_YLABEL_COUNT = "Count"
    PLOT_YLABEL_SOC = "SoC"
    PLOT_YLABEL_CURTAIL = "Curtailment (%)"
    PLOT_YLABEL_STATUS = "Vehicle Status"
    PLOT_YLABEL_PRICE = "Price ($/kWh)"
    PLOT_XLABEL_TIME = "Time"
    PLOT_FCS_TITLE = "Fast Charging Station: {0}"
    PLOT_SCS_TITLE = "Slow Charging Station: {0}"
    PLOT_BUS_TOTAL = "Total Bus Load"
    PLOT_GEN_TOTAL = "Total Generator"
    PLOT_LINE = "Line: {0}"
    PLOT_GEN = "Generator: {0}"
    PLOT_BUS = "Bus: {0}"
    PLOT_PVW = "PV or Wind Turbine: {0}"
    PLOT_ESS = "ESS: {0}"

    PLOT_TOOL_CLEARING = "Clearing {0}..."
    PLOT_TOOL_PLOTTING = "Plotting {0}..."
    PLOT_TOOL_EMPTY = "No plots to generate. Please check your options."
    PLOT_TOOL_INDIR_NOT_FOUND = "Input directory {0} not found."
    PLOT_TOOL_UNKNOWN_ARG = "Unknown argument: {0}"
    PLOT_TOOL_NO_RESULTS = "No results found in {0}. Try using -r for recursive plotting."
    PLOT_TOOL_NO_RESULTS_RECURSIVE = "No results found in {0} and its subdirectories."

    CONVERT_ERROR_MISSING_PATHS = "Error: Please provide input and output directory paths by '-i' and '-o'."
    CSQUERY_KEY_REQUIRED = "Please provide an AMap key in command line with '--key'"
    BAD_TRIP_OD = "Trip start node must be the same as the previous trip's end node, but got {0} after {1} for vehicle {2}'s trip {3}"
    BAD_TRIP_DEPART_TIME = "Trip departure time must be in ascending order, but got {0} after {1} for vehicle {2}'s trip {3}"
    PARA_WORLDS = "Paraworlds created. Number of sub-worlds: {0}."
    SINGLE_WORLD = "Single world created."
    ROUTE_ALGO_NOT_SUPPORTED = "Route calculation algorithm '{0}' is not supported. Candidates are: dijkstra, astar"
    VEH_NOT_FOUND = "Vehicle {0} not found in the simulator."
    VEH_HAS_NO_LINK = "Vehicle {0} has no link assigned."
    NO_AVAILABLE_FCS = "No FCS is available at this time, please check the configuration."
    EV_LOAD_ERROR = "Failed to load EV file {0}"
    NODE_EXISTS = "Node {0} already exists."
    NODE_NOT_FOUND = "Node {0} does not exist."
    EDGE_EXISTS = "Edge {0} already exists."
    EDGE_NOT_FOUND = "Edge {0} does not exist."
    INVALID_SUMO_NETWORK = "Invalid SUMO network: {0}"
    UNKNOWN_NET_FORMAT = "Unknown network format: {0}. Candidates: raw, sumo, auto."
    GIL_NOT_DISABLED = "Warning: ParaWorlds requires Python to be built with GIL disabled. Falling back to SingleWorld."
    NO_GEO_PROJ = "Network does not provide geo-projection"
    SIMULATION_MAY_STALL = "Simulation may stall: average speed < 0.001 m/s at time {0}."
    PY_VERSION_MISMATCH_TI = "Python version mismatch for TrafficInst: Expect {0}, got {1}."
    PY_VERSION_MISMATCH_PLG = "Incompatible Python version for plugin states: saved {0}, current {1}."
    PICKLER_MISMATCH_TI = "Pickler mismatch for TrafficInst: Expect {0}, got {1}."
    PICKLER_MISMATCH_PLG = "Incompatible pickler for plugin states: saved {0}, current {1}."
    SAVED_STATE_COPIED = "Saved state files have copied to the project folder."
    CASE_FILE_COPIED = "Case files have been copied to the result folder."
    NO_HOST_EXISTS = "Not working in multiprocessing mode. No host exists."
    ALT_COMMAND_NOT_SUPPORTED = "Cannot use alt_command when initial_state is specified."
    INVALID_PLUGIN_STATES = "Invalid plugin states in saved state: {0}."

    @staticmethod
    def format(item:str, **kwargs):
        fmt = getattr(Lang, item)
        assert isinstance(fmt, str), f"Invalid item {item}"
        return fmt.format(**kwargs)

    @staticmethod
    def load(lang:str)->bool:
        if lang == "en_US" or lang == "en":
            lc = en_Lang
        else:
            try:
                m = importlib.import_module(f"v2simux.locale.{lang}")
            except ImportError:
                return False
            if not hasattr(m, "_locale"):
                raise ValueError(f"Invalid language {lang}")
            lc = m._locale
        for key, val in lc.__dict__.items():
            assert isinstance(key,str)
            if key.startswith("__") and key.endswith("__"): continue
            if isinstance(val, str):
                if hasattr(Lang, key):
                    setattr(Lang, key, val)
                else:
                    print(f"Unknown key {key}")
        return True
    
    @staticmethod
    def load_default():
        locale_code = LangConfig.GetLangCode()
        if "en" in locale_code:
            return
        if Lang.load(locale_code):
            return
        Lang.load(locale_code.split("_")[0])

en_Lang = copy.deepcopy(Lang)
Lang.load_default()


from itertools import chain, repeat
from feasytools import LangLib
from .base import *

FILE_FCS = "fcs"
FILE_SCS = "scs"
CS_ATTRIB = ["cnt","c","d","v2g","pb","ps"]

_L = LangLib(["en", "zh_CN"])
_L.SetLangLib("en",
    FCS = "FCS",
    SCS = "SCS",
)
_L.SetLangLib("zh_CN",
    FCS = "快充站",
    SCS = "慢充站",
)

class StaFCS(StaBase):
    def __init__(self,path:str,tinst:TrafficInst,plugins:Dict[str,PluginBase]):
        head = cross_list(tinst.FCSList.get_CS_names(),["cnt","c","pb"])
        super().__init__(FILE_FCS,path,head,tinst,plugins)

    @staticmethod
    def GetLocalizedName() -> str:
        return _L("FCS")
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return []
    
    def GetData(self,inst:TrafficInst,plugins:Dict[str,PluginBase])->Iterable[Any]:
        t = inst.current_time
        IL = inst.FCSList
        cnt = (cs.__len__() for cs in IL)
        Pc = (cs._cload * 3600 for cs in IL)
        pb = (cs.pbuy(t) for cs in IL)
        return chain(cnt, Pc, pb)

class StaSCS(StaBase):
    def __init__(self,path:str,tinst:TrafficInst,plugins:Dict[str,PluginBase]):
        head = cross_list(tinst.SCSList.get_CS_names(),CS_ATTRIB)
        super().__init__(FILE_SCS,path,head,tinst,plugins)
        self.L = len(tinst.SCSList)
        self.supv2g = self.L > 0 and tinst.SCSList[0].supports_V2G
        self.hasv2g = "v2g" in plugins

    @staticmethod
    def GetLocalizedName() -> str:
        return _L("SCS")
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return []
    
    
    def GetData(self,inst:TrafficInst,plugins:Dict[str,PluginBase])->Iterable[Any]:
        L = self.L
        IL = inst.SCSList
        t = inst.current_time
        cnt = (cs.__len__() for cs in IL)
        Pc = (cs._cload * 3600 for cs in IL) # Performance problem: do not call property
        pb = (cs._pbuy(t) for cs in IL)
        if self.supv2g:
            v2g = self.hasv2g and plugins["v2g"].IsOnline(t)
            ps = (cs._psell(t) for cs in IL) # type: ignore
            Pd = (cs._dload * 3600 for cs in IL) if v2g else repeat(0, L)
            Pv2g = (cs._cur_v2g_cap * 3600 for cs in IL) if v2g else repeat(0, L)
        else:
            ps = repeat(0,L)
            Pd = repeat(0,L)
            Pv2g = repeat(0,L)
        return chain(cnt,Pc,Pd,Pv2g,pb,ps)


from v2simux_gui.common import *

from v2simux import MsgPack
import multiprocessing as mp
import sys, time


ITEM_NONE = "none"
_L = LangLib.Load(__file__)


class RedirectStdout:
    def __init__(self, q:mp.Queue, id:int):
        self.q = q
        self.ln = id

    def write(self, text):
        self.q.put((self.ln, text))

    def flush(self):
        pass


def work(root:str, par:Dict[str, str], alt:Dict[str, str], out:str, recv:RedirectStdout):
    sys.stdout = recv
    import sim_single
    par.update({"d":root, "od":out})
    st_time = time.time()
    sim_single.work(par, recv.ln, recv.q, alt)
    recv.q.put_nowait(MsgPack(recv.ln, f"done:{time.time()-st_time:.2f}"))


__all__ = ["RedirectStdout", "ITEM_NONE", "mp", "_L", "work"]


from v2simux_gui.common import *

_L = LangLib.Load(__file__)

class SelectItemDialog(Toplevel):
    def __init__(self, items:List[List[Any]], title:str, columns:List[Union[str, Tuple[str, str]]]):
        super().__init__()
        self.title(title)
        self.geometry("500x300")
        self.__dkt:Dict[str, Any] = {}
        self.selected_item = None

        col_id = []
        col_name = []
        for col in columns:
            if isinstance(col, str):
                col_id.append(col)
                col_name.append(col.capitalize())
            else:
                col_id.append(col[0])
                col_name.append(col[1])
        tree = Treeview(self, columns=col_id, show="headings", selectmode="browse")
        for i, n in zip(col_id, col_name):
            tree.heading(i, text=n)

        for item in items:
            idx = tree.insert("", "end", values=item)
            self.__dkt[idx] = item
        tree.pack(fill="both", expand=True, padx=10, pady=10)
        tree.bind("<<TreeviewSelect>>", self.on_select)
        self.tree = tree

        btn = Button(self, text=_L("CONFIRM"), command=self.confirm_selection)
        btn.pack(pady=5)

    def on_select(self, event):
        selected = self.tree.selection()
        if selected:
            self.selected_item = self.__dkt[selected[0]]

    def confirm_selection(self):
        if self.selected_item is not None:
            self.destroy()
        else:
            MB.showwarning(_L("WARNING"), _L("HINT_SELECT_ITEM"))


__all__ = ["SelectItemDialog"]


from v2simux_gui.common import *

from v2simux.trafficgen.misc import *

class EditDesc:
    def __init__(self, typename:type):
        self._t = typename.__name__
        self._desc:Dict[str, str] = {}
        self._text:Dict[str, str] = {}
        self._dtype:Dict[str, type] = {}
        self._em:Dict[str, EditMode] = {}
        self._em_kwargs:Dict[str, Dict[str, Any]] = {}
        self._onchanged:Dict[str, Optional[Callable[[Any, Any], None]]] = {}
    
    def add(self, key:str, show:str, dtype:type, desc:str, 
            edit_mode:EditMode, onchanged = None, **kwargs):
        self._desc[key] = desc
        self._text[key] = show
        self._dtype[key] = dtype
        self._em[key] = edit_mode
        self._em_kwargs[key] = kwargs
        self._onchanged[key] = onchanged
        return self
    
    @staticmethod
    def create(typename:type, default_edit_mode:EditMode):
        return EditDesc(typename)


class EditDescGroup:
    def __init__(self, EditDescs:Iterable[EditDesc]):
        self._eds = {ed._t:ed for ed in EditDescs}
    
    def get(self, inst:Any) -> EditDesc:
        typename = type(inst).__name__
        if typename not in self._eds:
            raise KeyError(f"Type {typename} not found in EditDescGroup")
        return self._eds[typename]
    
__all__ = ["EditDesc", "EditDescGroup"]


'''
External Plugin Example: 
  Please put the external plugin in the external_plugins folder, 
  and the plugin file name is "plugin_name.py"
'''
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Iterable, Any, Tuple, List, Dict
from feasytools import LangLib
from v2simux import TrafficInst
from v2simux.plugins import *
from v2simux.statistics import *

_L = LangLib.LoadFor(__file__)

class DemoExternalPlugin(PluginBase):
    @property
    def Description(self)->str:
        return _L["DESCRIPTION"]
    
    def Init(self, elem:ET.Element, inst:TrafficInst, work_dir:Path, plg_deps:List[PluginBase]) -> object:
        '''
        Add plugin initialization code here, return:
            Return value when the plugin is offline
        '''
        self.SetPreStep(self.Work) # Indicate that the Work function should be called in PreStep (i.e. before SUMO simulation step)
        return None

    def Work(self, _t:int, /, sta:PluginStatus)->Tuple[bool,None]:
        '''The execution function of the plugin at time _t'''
        raise NotImplementedError

class DemoStatisticItem(StaBase):
    @property
    def Description(self)->str:
        return _L["DESCRIPTION"]
    
    def __init__(self, name:str, path:str, items:List[str], tinst:TrafficInst, 
            plugins:Dict[str,PluginBase], precision:Dict[str, int]={}, compress:bool=True):
        super().__init__(name, path, items, tinst, plugins, precision, compress)
        raise NotImplementedError

    @staticmethod
    def GetLocalizedName() -> str:
        '''Get Localized Name'''
        return _L["STA_NAME"]
    
    @staticmethod
    def GetPluginDependency() -> List[str]:
        '''Get Plugin Dependency'''
        return ["demo"] # Example: This statistic item depends on the "demo" plugin
    
    def GetData(self, inst:TrafficInst, plugins:Dict[str,PluginBase]) -> Iterable[Any]: 
        '''Get Data'''
        raise NotImplementedError

'''
Set export variables
  plugin_exports = (Plugin name, Plugin class, Plugin dependency list(can be empty))
  sta_exports = (Statistic item name, Statistic item class)
If you don't export the statistic item, please don't set sta_exports
'''

plugin_exports = ("demo", DemoExternalPlugin, ["pdn"])
sta_exports = ("demo", DemoStatisticItem)


"""
Scenario reader and writer using toml.
"""

import functools
import inspect
from collections import defaultdict as ddict
import dill as pickle

def demand_info_record(func):
    """
    A decorator to record arguments of `World.addVehicle`, `World.adddemand`, etc used in scenario definition in World object.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # 関数名を取得
        func_name = func.__name__
        
        # 引数の情報を取得
        bound_args = inspect.signature(func).bind(self, *args, **kwargs)
        bound_args.apply_defaults()
        
        # self引数を除外
        arg_dict = dict(list(bound_args.arguments.items())[1:])
        
        # direct_call=Trueが含まれている場合のみ保存
        if arg_dict.get('direct_call') == True:
            
            # direct_callを除いた引数を保存
            arg_dict_to_save = {k: v for k, v in arg_dict.items() if k != 'direct_call'}
            #arg_dict_to_save["] = func_name
            
            # 関数名と引数をdict_func_argsに格納
            self.demand_info[func_name].append(arg_dict_to_save)
        
        # 元の関数を呼び出す
        return func(self, *args, **kwargs)
    
    return wrapper

def instance_to_arg_dict(instance):
    """
    Convert an instance's attributes to a dictionary of constructor arguments.

    This function inspects an object instance and creates a dictionary of
    arguments that could be used to recreate the instance. It uses the
    signature of the class's __init__ method to determine which attributes
    to include.

    Parameters
    ----------
    instance : object
        The instance to convert to a dictionary of arguments.

    Returns
    -------
    dict
        A dictionary where keys are parameter names and values are the 
        corresponding attribute values from the instance.

    Notes
    -----
    - Parameters with default values are omitted if the instance's value
      matches the default.
    - The 'self' parameter is always ignored.
    - The 'W' parameter, if present, is removed from the resulting dictionary.
    - Attributes of type 'Node' or 'Link' are replaced with their 'name' attribute.
    """
    
    # クラスの__init__メソッドのシグネチャを取得
    signature = inspect.signature(instance.__class__.__init__)
    
    # インスタンス変数を取得
    instance_vars = vars(instance)
    
    # 引数リストを作成
    args = {}
    for param_name, param in signature.parameters.items():
        if param_name == 'self':
            continue
        
        value = instance_vars.get(param_name, param.default)
      
        if value == param.default:
            continue
        
        args[param_name] = value
        for param in list(args):
            if param == "W":
                del args[param]            
            elif str(type(args[param])) == "<class 'uxsim.uxsim.Node'>":
                args[param] = args[param].name
            elif str(type(args[param])) == "<class 'uxsim.uxsim.Link'>":
                args[param] = args[param].name

    return args

def save_scenario(W, fname, network=True, demand=True):
    """
    Save scenario data to a file.

    This function saves various components of a scenario, including meta data,
    network information (nodes and links), and demand information, to a file 
    using pickle serialization.

    Parameters
    ----------
    W : World
        The world to be saved
    fname : str
        The file name or path where the scenario data will be saved.
    network : bool, optional
        If True, save network information (nodes and links). Default is True.
    demand : bool, optional
        If True, save demand information. Default is True.

    Notes
    -----
    The function uses the pickle module to serialize the data. The saved file
    will be in binary format.

    The saved data structure is a dictionary containing:
    - 'meta_data': Meta data from W.meta_data
    - 'Nodes' and 'Links': Network information (if network=True)
    - Demand information from W.demand_info (if demand=True)
    """

    out = {"meta_data": W.meta_data}

    if network:
        nodes = []
        for n in W.NODES:
            nodes.append(instance_to_arg_dict(n))
        links = []
        for l in W.LINKS:
            links.append(instance_to_arg_dict(l))
        network = {
            "Nodes": nodes,
            "Links": links
        }
        out = out | network
    if demand:
        out = out | dict(W.demand_info)
    
    # from pprint import pprint
    # pprint(out)
    with open(fname, "wb") as f:
        pickle.dump(out, f)
    

def load_scenario(W, fname, network=True, demand=True):
    """
    Load scenario data from a file.

    This function loads various components of a scenario, including meta data,
    network information (nodes and links), and demand information, from a file 
    created by the save_scenario function.

    Parameters
    ----------
    W : World
        The world to which the scenario data will be loaded.
    fname : str
        The file name or path from which the scenario data will be loaded.
    network : bool, optional
        If True, load network information (nodes and links). Default is True.
    demand : bool, optional
        If True, load demand information. Default is True.
        print_meta_data : bool, optional
            If True, print the meta data of the scenario file, such as licence info. Default is True.
    """

    with open(fname, "rb") as f:
        dat = pickle.load(f)
    
    # from pprint import pprint
    # pprint(dat)

    W.print(f"loading scenario from '{fname}'")
    if dat["meta_data"]:
        if type(dat["meta_data"]) is dict:
            for key in dat["meta_data"]:
                W.print("", key, ":", dat["meta_data"][key])
        else:
            W.print("", dat["meta_data"])


    if network:
        for n in dat["Nodes"]:
            W.addNode(**n)
        for l in dat["Links"]:
            W.addLink(**l)
        W.print(" Number of loaded nodes:", len(dat["Nodes"]))
        W.print(" Number of loaded links:", len(dat["Links"]))
    
    if demand:
        for demand_type in dat:
            if demand_type == 'adddemand':
                for dem in dat[demand_type]:
                    W.adddemand(**dem)
                W.print(f" Number of loaded `{demand_type}`s:", len( dat[demand_type]))
            elif demand_type == 'adddemand_point2point':
                for dem in dat[demand_type]:
                    W.adddemand_point2point(**dem)
                W.print(f" Number of loaded `{demand_type}`s:", len( dat[demand_type]))
            elif demand_type == 'adddemand_area2area':
                for dem in dat[demand_type]:
                    W.adddemand_area2area(**dem)
                W.print(f" Number of loaded `{demand_type}`s:", len( dat[demand_type]))
            elif demand_type == 'adddemand_nodes2nodes':
                for dem in dat[demand_type]:
                    W.adddemand_nodes2nodes(**dem)
                W.print(f" Number of loaded `{demand_type}`s:", len( dat[demand_type]))
            elif demand_type == 'adddemand_area2area2':
                for dem in dat[demand_type]:
                    W.adddemand_area2area2(**dem)
                W.print(f" Number of loaded `{demand_type}`s:", len( dat[demand_type]))
            elif demand_type == 'adddemand_nodes2nodes2':
                for dem in dat[demand_type]:
                    W.adddemand_nodes2nodes2(**dem)
                W.print(f" Number of loaded `{demand_type}`s:", len( dat[demand_type]))
            else:
                pass
        







from v2simux_gui.common import *
from v2simux_gui.langhelper import *

import os
import sys
import time
import traceback
import xml.etree.ElementTree as ET
from pathlib import Path
from fpowerkit import Grid as PowerGrid
from feasytools import RangeList, PDUniform
from v2simux import (
    FileDetectResult, TrafficGenerator, V2SimConfig, DetectFiles,
    ReadXML, PluginBase, PluginPDN, RoadNet, csQuery, TripsGenMode, RoutingCacheMode
)
from .utils import *
from .loadingbox import LoadingBox
from .plugin import PluginEditor
from .cseditor import CSEditorGUI
from .cscsveditor import CSCSVEditor
from .controls import LogItemPad, PropertyPanel, PDFuncEditor, ALWAYS_ONLINE, NetworkPanel, OAfter


_L = LangLib.Load(__file__)
DEFAULT_GRID = '<grid Sb="1MVA" Ub="10.0kV" model="ieee33" fixed-load="false" grid-repeat="1" load-repeat="8" />'
LOAD_FCS = "Fast CS"
LOAD_SCS = "Slow CS"
LOAD_NET = "Network"
LOAD_CSCSV = "CS CSV"
LOAD_PLG = "Plugins"
LOAD_GEN = "Instance"
    

class MainBox(Tk):
    def __OnPluginEnabledSet(self, itm:Tuple[Any,...]=(), v:str=""):
        plgs = self.sim_plglist.GetEnabledPlugins()
        self.sim_statistic.check_by_enabled_plugins(plgs)
    
    def __init__(self, to_open:str = ""):
        super().__init__()
        self._Q = EventQueue(self)
        
        def proc_exception(e: Optional[Exception] = None):
            if e:
                self.setStatus(f"Error: {e}")
                showerr(f"Error: {e}")
            else:
                self.setStatus(_L["STA_READY"])
       
        def on_CSGendone(ctl: CSEditorGUI, e: Optional[Exception] = None,
                    warns:List[Tuple] = [], far_cnt = 0, scc_cnt = 0):
            ctl.btn_regen.config(state=NORMAL)
            proc_exception(e)

            if len(warns) == 0: return
            
            with open("CS_generation_warnings.log", "w") as fh:
                for ln in warns:
                    if ln[0] == "far_poly":
                        fh.write(f"A polygon (center: {ln[1]:.1f},{ln[2]:.1f}) is far away ({ln[3]:.1f}m) from the road network.\n")
                    elif ln[0] == "scc_poly":
                        fh.write(f"A polygon (center: {ln[1]:.1f},{ln[2]:.1f}) is not neighbouring to an edge in the largest SCC and allowing passengers.\n")
                    elif ln[0] == "far_down":
                        fh.write(f"Point {ln[1]},{ln[2]} (XY: {ln[3]:.1f},{ln[4]:.1f}) is far away ({ln[5]:.1f}m) from the road network.\n")
                    elif ln[0] == "scc_down":
                        fh.write(f"The nearest edge of point {ln[1]},{ln[2]} (XY: {ln[3]:.1f},{ln[4]:.1f}) is not in the max SCC which allows passengers.\n")
                    elif ln[0] == "scc_name":
                        fh.write(f"Edge {ln[1]} disallows passenger vehicles.")

            text = "Some warnings have been written in CS_generation_warnings.log:"
            if far_cnt: 
                text += f"{far_cnt} CS(s) are abondoned since their distance to the nearest edge is greater than 200m."
            if scc_cnt: 
                if text != "": text += "\n"
                text += f"{far_cnt} CS(s) are abondoned since they are not in the largest strongly connected component or disallow passenger vehicles."
            showwarn(text)
        
        self._Q.register("CSGenDone", on_CSGendone)

        def on_VehGenDone(e: Optional[Exception] = None):
            self.btn_genveh.config(state = NORMAL)
            proc_exception(e)
        
        self._Q.register("VehGenDone", on_VehGenDone)

        def on_CSCSVDownloadDone(e: Optional[Exception] = None):
            proc_exception(e)
        
        self._Q.register("CSCSVDownloadDone", on_CSCSVDownloadDone)

        def on_TrafficGenLoaded():
            self._ldfrm.setText(LOAD_GEN, _L['DONE'])
        self._Q.register("TrafficGenLoaded", on_TrafficGenLoaded)

        
        self._Q.register("cvnetloaded", self.on_cvnet_loaded)

        self.folder:str = to_open
        self.state:Optional[FileDetectResult] = None
        self.tg:Optional[TrafficGenerator] = None
        self._win()

        self.menu = Menu(self)
        self.menuFile = Menu(self.menu, tearoff=False)
        self.menu.add_cascade(label=_L["MENU_PROJ"], menu=self.menuFile)
        self.menuFile.add_command(label=_L["MENU_OPEN"], command=self.openFolder, accelerator='Ctrl+O')
        self.bind("<Control-o>", lambda e: self.openFolder())
        self.menuFile.add_command(label=_L["MENU_SAVEALL"], command=self.save, accelerator="Ctrl+S")
        self.bind("<Control-s>", lambda e: self.save())
        self.menuFile.add_separator()
        self.menuFile.add_command(label=_L["MENU_EXIT"], command=self.onDestroy, accelerator='Ctrl+Q')
        self.bind("<Control-q>", lambda e: self.onDestroy())
        add_lang_menu(self.menu)
        self.config(menu=self.menu)
        self.rowconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)

        self.panel_info = Frame(self, borderwidth=1, relief="solid")
        self.panel_info.grid(row=0, column=0, padx=3, pady=3, sticky="nsew")

        self.lb_infotitle = Label(self.panel_info, text = _L["BAR_PROJINFO"], background="white")
        self.lb_infotitle.grid(row=0, column=0, columnspan=2, padx=3, pady=3, sticky="nsew")

        self.lb_fcs_indicatif = Label(self.panel_info, text = _L["BAR_FCS"])
        self.lb_fcs_indicatif.grid(row=1, column=0, padx=3, pady=3)
        self.lb_fcs = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_fcs.grid(row=1, column=1, padx=3, pady=3)

        self.lb_scs_indicatif = Label(self.panel_info, text = _L["BAR_SCS"])
        self.lb_scs_indicatif.grid(row=2, column=0, padx=3, pady=3)
        self.lb_scs = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_scs.grid(row=2, column=1, padx=3, pady=3)

        self.lb_grid_indicatif = Label(self.panel_info, text = _L["BAR_GRID"])
        self.lb_grid_indicatif.grid(row=3, column=0, padx=3, pady=3)
        self.lb_grid = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_grid.grid(row=3, column=1, padx=3, pady=3)

        self.lb_net_indicatif = Label(self.panel_info, text = _L["BAR_RNET"])
        self.lb_net_indicatif.grid(row=4, column=0, padx=3, pady=3)
        self.lb_net = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_net.grid(row=4, column=1, padx=3, pady=3)

        self.lb_veh_indicatif = Label(self.panel_info, text = _L["BAR_VEH"])
        self.lb_veh_indicatif.grid(row=5, column=0, padx=3, pady=3)
        self.lb_veh = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_veh.grid(row=5, column=1, padx=3, pady=3)

        self.lb_plg_indicatif = Label(self.panel_info, text = _L["BAR_PLG"])
        self.lb_plg_indicatif.grid(row=6, column=0, padx=3, pady=3)
        self.lb_plg = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_plg.grid(row=6, column=1, padx=3, pady=3)

        self.lb_cscsv_indicatif = Label(self.panel_info, text = _L["BAR_CSCSV"])
        self.lb_cscsv_indicatif.grid(row=7, column=0, padx=3, pady=3)
        self.lb_cscsv = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_cscsv.grid(row=7, column=1, padx=3, pady=3)

        self.lb_node_type_indicatif = Label(self.panel_info, text = _L["BAR_NODETYPE"])
        self.lb_node_type_indicatif.grid(row=8, column=0, padx=3, pady=3)
        self.lb_node_type = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_node_type.grid(row=8, column=1, padx=3, pady=3)

        self.lb_py_indicatif = Label(self.panel_info, text = _L["BAR_ADDON"])
        self.lb_py_indicatif.grid(row=9, column=0, padx=3, pady=3)
        self.lb_py = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_py.grid(row=9, column=1, padx=3, pady=3)

        self.lb_osm_indicatif = Label(self.panel_info, text = _L["BAR_OSM"])
        self.lb_osm_indicatif.grid(row=10, column=0, padx=3, pady=3)
        self.lb_osm = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_osm.grid(row=10, column=1, padx=3, pady=3)

        self.lb_poly_indicatif = Label(self.panel_info, text = _L["BAR_POLY"])
        self.lb_poly_indicatif.grid(row=11, column=0, padx=3, pady=3)
        self.lb_poly = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_poly.grid(row=11, column=1, padx=3, pady=3)

        self.lb_poi_indicatif = Label(self.panel_info, text = _L["BAR_POI"])
        self.lb_poi_indicatif.grid(row=12, column=0, padx=3, pady=3)
        self.lb_poi = Label(self.panel_info, text = _L["BAR_NONE"])
        self.lb_poi.grid(row=12, column=1, padx=3, pady=3)

        self.tabs = Notebook(self)
        self.tabs.grid(row=0, column=1, padx=3, pady=3, sticky="nsew")

        self.tab_sim = Frame(self.tabs)
        self.sim_time = LabelFrame(self.tab_sim, text=_L["SIM_BASIC"])
        self.sim_time.pack(fill="x", expand=False)
        self.lb_start = Label(self.sim_time, text=_L["SIM_BEGT"])
        self.lb_start.grid(row=0, column=0, padx=3, pady=3, sticky="w")
        self.entry_start = Entry(self.sim_time)
        self.entry_start.insert(0, "0")
        self.entry_start.grid(row=0, column=1, padx=3, pady=3, sticky="w")

        self.lb_break = Label(self.sim_time, text=_L["SIM_BREAKT"])
        self.lb_break.grid(row=1, column=0, padx=3, pady=3, sticky="w")
        self.entry_break = Entry(self.sim_time)
        self.entry_break.insert(0, "172800")
        self.entry_break.grid(row=1, column=1, padx=3, pady=3, sticky="w")

        self.lb_end = Label(self.sim_time, text=_L["SIM_ENDT"])
        self.lb_end.grid(row=2, column=0, padx=3, pady=3, sticky="w")
        self.entry_end = Entry(self.sim_time)
        self.entry_end.insert(0, "172800")
        self.entry_end.grid(row=2, column=1, padx=3, pady=3, sticky="w")

        self.lb_step = Label(self.sim_time, text=_L["SIM_STEP"])
        self.lb_step.grid(row=3, column=0, padx=3, pady=3, sticky="w")
        self.entry_step = Entry(self.sim_time)
        self.entry_step.insert(0, "10")
        self.entry_step.grid(row=3, column=1, padx=3, pady=3, sticky="w")

        self.lb_seed = Label(self.sim_time, text=_L["SIM_SEED"])
        self.lb_seed.grid(row=4, column=0, padx=3, pady=3, sticky="w")
        self.entry_seed = Entry(self.sim_time)
        self.entry_seed.insert(0, "0")
        self.entry_seed.grid(row=4, column=1, padx=3, pady=3, sticky="w")

        self.sim_state_load_panel = Frame(self.sim_time)
        self.sim_state_load_panel.grid(row=0, column=2, padx=0, pady=0, sticky="w")

        self.sim_lb_state_load = Label(self.sim_state_load_panel, text=_L["SIM_STATE_LOAD_OPTIONS"])
        self.sim_lb_state_load.grid(row=0, column=0, padx=3, pady=3, sticky="w")

        self.sim_load_state = IntVar(self, 0)

        self.sim_cb_no_load_state = Radiobutton(self.sim_state_load_panel, text=_L["SIM_NO_LOAD_STATE"], 
            value=0, variable=self.sim_load_state, command=self.on_load_state_changed)
        self.sim_cb_no_load_state.grid(row=0, column=1, padx=3, pady=3, sticky="w")
        
        self.sim_cb_load_last_state = Radiobutton(self.sim_state_load_panel, text=_L["SIM_LOAD_LAST_STATE"],
            value=1, variable=self.sim_load_state, command=self.on_load_state_changed)
        self.sim_cb_load_last_state.grid(row=0, column=2, padx=3, pady=3, sticky="w")

        self.sim_cb_load_saved_state = Radiobutton(self.sim_state_load_panel, text=_L["SIM_LOAD_SAVED_STATE"],
            value=2, variable=self.sim_load_state, command=self.on_load_state_changed)
        self.sim_cb_load_saved_state.grid(row=0, column=3, padx=3, pady=3, sticky="w")

        self.sim_state_save_panel = Frame(self.sim_time)
        self.sim_state_save_panel.grid(row=1, column=2, padx=0, pady=0, sticky="w")
        
        self.sim_lb_state_save = Label(self.sim_state_save_panel, text=_L["SIM_STATE_SAVE_OPTIONS"])
        self.sim_lb_state_save.grid(row=0, column=0, padx=3, pady=3, sticky="w")

        self.sim_save_on_abort = BooleanVar(self, False)
        self.sim_cb_save_on_abort = Checkbutton(self.sim_state_save_panel, text=_L["SIM_SAVE_ON_ABORT"], variable=self.sim_save_on_abort)
        self.sim_cb_save_on_abort.grid(row=0, column=1, padx=3, pady=3, sticky="w")

        self.sim_save_on_finish = BooleanVar(self, False)
        self.sim_cb_save_on_finish = Checkbutton(self.sim_state_save_panel, text=_L["SIM_SAVE_ON_FINISH"], variable=self.sim_save_on_finish)
        self.sim_cb_save_on_finish.grid(row=0, column=2, padx=3, pady=3, sticky="w")

        self.sim_copy_state = BooleanVar(self, False)
        self.sim_cb_copy_state = Checkbutton(self.sim_state_save_panel, text=_L["SIM_COPY_STATE"], variable=self.sim_copy_state)
        self.sim_cb_copy_state.grid(row=0, column=3, padx=3, pady=3, sticky="w")

        self.sim_ux_options_panel = Frame(self.sim_time)
        self.sim_ux_options_panel.grid(row=2, column=2, padx=0, pady=0, sticky="w")
        
        self.sim_no_parallel = BooleanVar(self, False)
        self.sim_cb_no_parallel = Checkbutton(self.sim_ux_options_panel, text=_L["SIM_NO_PARALLEL"], variable=self.sim_no_parallel)
        self.sim_cb_no_parallel.grid(row=0, column=0, padx=3, pady=3, sticky="w")

        self.sim_show_uxsim_info = BooleanVar(self, False)
        self.sim_cb_show_uxsim_info = Checkbutton(self.sim_ux_options_panel, text=_L["SIM_SHOW_UXSIM_INFO"], variable=self.sim_show_uxsim_info)
        self.sim_cb_show_uxsim_info.grid(row=0, column=1, padx=3, pady=3, sticky="w")

        self.sim_randomize_traffic = BooleanVar(self, False)
        self.sim_cb_randomize_traffic = Checkbutton(self.sim_ux_options_panel, text=_L["SIM_RANDOMIZE_TRAFFIC"], variable=self.sim_randomize_traffic)
        self.sim_cb_randomize_traffic.grid(row=0, column=2, padx=3, pady=3, sticky="w")

        self.sim_algo_panel = Frame(self.sim_time)
        self.sim_algo_panel.grid(row=4, column=2, padx=0, pady=0, sticky="w")

        self.ralgo = StringVar(self, "astar")
        self.lb_route_algo = Label(self.sim_algo_panel, text=_L["SIM_ROUTE_ALGO"])
        self.lb_route_algo.grid(row=0, column=0, padx=3, pady=3, sticky="w")
        self.combo_ralgo = Combobox(self.sim_algo_panel, textvariable=self.ralgo, values=["dijkstra", "astar"])
        self.combo_ralgo.grid(row=0, column=1, padx=3, pady=3, sticky="w")

        self.sim_plugins = LabelFrame(self.tab_sim, text=_L["SIM_PLUGIN"])
        self.sim_plglist = PluginEditor(self.sim_plugins, self.__OnPluginEnabledSet)
        self.sim_plglist.pack(fill="both", expand=True)
        self.sim_plugins.pack(fill="x", expand=False)
        self.sim_plglist.setOnSave(self.savePlugins())
        self.sim_plglist.AfterFunc = self.__OnPluginEnabledSet

        self.sim_statistic = LogItemPad(self.tab_sim, _L["SIM_STAT"],self.sim_plglist.sta_pool)
        self.sim_statistic["ev"] = False
        self.sim_statistic.pack(fill="x", expand=False)
        self.__OnPluginEnabledSet()

        self.sim_btn = Button(self.tab_sim, text=_L["SIM_START"], command=self.simulate)
        self.sim_btn.pack(anchor="w", padx=3, pady=3)
        self.tabs.add(self.tab_sim, text=_L["TAB_SIM"])

        self.tab_CsCsv = Frame(self.tabs)
        self.CsCsv_editor = CSCSVEditor(self.tab_CsCsv, self.CSCSVDownloadWorker)
        self.CsCsv_editor.pack(fill="both", expand=True)
        self.tabs.add(self.tab_CsCsv, text=_L["TAB_CSCSV"])

        self.tab_FCS = Frame(self.tabs)
        self.FCS_editor = CSEditorGUI(self.tab_FCS, self.generateCS, False)
        self.FCS_editor.pack(fill="both", expand=True)
        self.tabs.add(self.tab_FCS, text=_L["TAB_FCS"])

        self.tab_SCS = Frame(self.tabs)
        self.SCS_editor = CSEditorGUI(self.tab_SCS, self.generateCS, True)
        self.SCS_editor.pack(fill="both", expand=True)
        self.tabs.add(self.tab_SCS, text=_L["TAB_SCS"])

        self.tab_Net = Frame(self.tabs)
        self.cv_net = NetworkPanel(self.tab_Net)
        self.cv_net.pack(fill=BOTH, expand=True)
        self.panel_net = LabelFrame(self.tab_Net, text=_L["RNET_TITLE"])
        self.lb_gridsave = Label(self.panel_net, text=_L["NOT_OPEN"])
        self.lb_gridsave.pack(side='left',padx=3,pady=3, anchor='w')
        def on_saved_changed(saved:bool):
            if saved:
                self.lb_gridsave.config(text=_L["SAVED"],foreground="green")
            else:
                self.lb_gridsave.config(text=_L["UNSAVED"],foreground="red")
        self.cv_net.save_callback = on_saved_changed
        self.btn_savegrid = Button(self.panel_net, text=_L["SAVE_GRID"], command=self.save)
        self.btn_savegrid.pack(side='left',padx=3,pady=3, anchor='w')
        self.lb_puvalues = Label(self.panel_net, text=_L["PU_VALS"].format('Null','Null'))
        self.lb_puvalues.pack(side='left',padx=3,pady=3, anchor='w')
        self.btn_savenetfig = Button(self.panel_net, text=_L["RNET_SAVE"], command=self.netsave)
        self.btn_savenetfig.pack(side="right", padx=3, pady=3, anchor="e")
        self.btn_draw = Button(self.panel_net, text=_L["RNET_DRAW"], command=self.draw)
        self.btn_draw.pack(side="right", padx=3, pady=3, anchor="e")
        self.entry_Ledges = Entry(self.panel_net)
        self.entry_Ledges.pack(side="right", padx=3, pady=3, anchor="e")
        self.lb_Ledges = Label(self.panel_net, text=_L["RNET_EDGES"])
        self.lb_Ledges.pack(side="right", padx=3, pady=3, anchor="e")
        
        self.panel_net.pack(fill="x", expand=False, anchor="s")
        self.tabs.add(self.tab_Net, text=_L["TAB_RNET"])

        self.tab_Veh = Frame(self.tabs)
        self.fr_veh_basic = LabelFrame(self.tab_Veh,text=_L["VEH_BASIC"])
        self.fr_veh_basic.pack(fill="x", expand=False)
        self.lb_carcnt = Label(self.fr_veh_basic, text=_L["VEH_COUNT"])
        self.lb_carcnt.grid(row=0, column=0, padx=3, pady=3, sticky="w")
        self.entry_carcnt = Entry(self.fr_veh_basic)
        self.entry_carcnt.insert(0, "10000")
        self.entry_carcnt.grid(row=0, column=1, padx=3, pady=3, sticky="w")
        self.lb_daycnt = Label(self.fr_veh_basic, text=_L["VEH_DAY_COUNT"])
        self.lb_daycnt.grid(row=1, column=0, padx=3, pady=3, sticky="w")
        self.entry_daycnt = Entry(self.fr_veh_basic)
        self.entry_daycnt.insert(0, "7")
        self.entry_daycnt.grid(row=1, column=1, padx=3, pady=3, sticky="w")
        self.lb_v2gprop = Label(self.fr_veh_basic, text=_L["VEH_V2GPROP"])
        self.lb_v2gprop.grid(row=2, column=0, padx=3, pady=3, sticky="w")
        self.entry_v2gprop = Entry(self.fr_veh_basic)
        self.entry_v2gprop.insert(0, "1.00")
        self.entry_v2gprop.grid(row=2, column=1, padx=3, pady=3, sticky="w")
        self.lb_v2gprop_info = Label(self.fr_veh_basic, text=_L["VEH_V2GPROP_INFO"])
        self.lb_v2gprop_info.grid(row=2, column=2, padx=3, pady=3, sticky="w")
        self.lb_carseed = Label(self.fr_veh_basic, text=_L["VEH_SEED"])
        self.lb_carseed.grid(row=3, column=0, padx=3, pady=3, sticky="w")
        self.entry_carseed = Entry(self.fr_veh_basic)
        self.entry_carseed.insert(0, "0")
        self.entry_carseed.grid(row=3, column=1, padx=3, pady=3, sticky="w")

        self.veh_pars = PropertyPanel(self.tab_Veh, {
            "Omega":repr(PDUniform(5.0, 10.0)),
            "KRel":repr(PDUniform(1.0, 1.2)),
            "KSC":repr(PDUniform(0.4, 0.6)),
            "KFC":repr(PDUniform(0.2, 0.25)),
            "KV2G":repr(PDUniform(0.65, 0.75)),
        }, ConfigItemDict((
            ConfigItem("Omega", EditMode.PDFUNC, _L["VEH_OMEGA_DESC"]),
            ConfigItem("KRel", EditMode.PDFUNC, _L["VEH_KREL_DESC"]),
            ConfigItem("KSC", EditMode.PDFUNC, _L["VEH_KSC_DESC"]),
            ConfigItem("KFC", EditMode.PDFUNC, _L["VEH_KFC_DESC"]),
            ConfigItem("KV2G", EditMode.PDFUNC, _L["VEH_KV2G_DESC"]),
        )))
        self.veh_pars.pack(fill="x", expand=False, pady=10)

        self.veh_gen_src = IntVar(self, 0)
        self.fr_veh_src = LabelFrame(self.tab_Veh,text=_L["VEH_ODSRC"])
        self.fr_veh_src.pack(fill="x", expand=False)
        self.rb_veh_src0 = Radiobutton(self.fr_veh_src, text=_L["VEH_ODAUTO"], value=0, variable=self.veh_gen_src)
        self.rb_veh_src0.grid(row=0, column=0, padx=3, pady=3, sticky="w")
        self.rb_veh_src1 = Radiobutton(self.fr_veh_src, text=_L["VEH_ODTYPE"], value=1, variable=self.veh_gen_src)
        self.rb_veh_src1.grid(row=1, column=0, padx=3, pady=3, sticky="w")
        self.rb_veh_src2 = Radiobutton(self.fr_veh_src, text=_L["VEH_ODPOLY"], value=2, variable=self.veh_gen_src)
        self.rb_veh_src2.grid(row=2, column=0, padx=3, pady=3, sticky="w")

        self.veh_route_cache = IntVar(self, 0)
        self.fr_veh_route_cache = LabelFrame(self.tab_Veh,text=_L["VEH_ROUTE_CACHE"])
        # self.fr_veh_route_cache.pack(fill="x", expand=False)
        self.rb_veh_route_cache0 = Radiobutton(self.fr_veh_route_cache, text=_L["VEH_ROUTE_NO_CACHE"], value=0, variable=self.veh_route_cache)
        self.rb_veh_route_cache0.grid(row=0, column=0, padx=3, pady=3, sticky="w")
        self.rb_veh_route_cache1 = Radiobutton(self.fr_veh_route_cache, text=_L["VEH_ROUTE_RUNTIME_CACHE"], value=1, variable=self.veh_route_cache)
        self.rb_veh_route_cache1.grid(row=1, column=0, padx=3, pady=3, sticky="w")
        self.rb_veh_route_cache2 = Radiobutton(self.fr_veh_route_cache, text=_L["VEH_ROUTE_STATIC_CACHE"], value=2, variable=self.veh_route_cache)
        self.rb_veh_route_cache2.grid(row=2, column=0, padx=3, pady=3, sticky="w")

        self.btn_genveh = Button(self.tab_Veh, text=_L["VEH_GEN"], command=self.generateVeh)
        self.btn_genveh.pack(anchor="w")
        self.tabs.add(self.tab_Veh, text=_L["TAB_VEH"])

        self.sbar = Label(self, text=_L["STA_READY"], anchor="w")
        self.sbar.grid(row=1, column=0, columnspan=2, sticky="ew")
        self.protocol("WM_DELETE_WINDOW", self.onDestroy)

        if self.folder != "":
            self.after(200, self._load)
    
    def on_load_state_changed(self):
        val = self.sim_load_state.get()
        s = NORMAL if val == 0 else DISABLED
        self.entry_start.config(state=s)
        # self.entry_break.config(state=s) # break time is always editable
        self.entry_end.config(state=s)
        self.entry_step.config(state=s)
        self.entry_seed.config(state=s)
        self.combo_ralgo.config(state=s)
        self.sim_cb_no_parallel.config(state=s)
        self.sim_cb_show_uxsim_info.config(state=s)
    
    def netsave(self):
        ret = filedialog.asksaveasfilename(
            defaultextension=".eps",
            filetypes=[
                (_L["EXT_EPS"],".eps"),             
            ]
        )
        if ret == "": return
        try:
            self.cv_net.savefig(ret)
        except RuntimeError:
            showerr(_L["RNET_SAVE_ERR"])
            return
        self.setStatus("Figure saved")

    def veh_par_edit(self, var:StringVar):
        def _f():
            e = PDFuncEditor(var)
            e.wait_window()
        return _f

    @property
    def saved(self):
        return self.sim_plglist.saved and self.FCS_editor.saved and self.SCS_editor.saved and self.cv_net.saved
    
    def save(self):
        if not self.sim_plglist.saved: self.sim_plglist.save()
        if not self.FCS_editor.saved: self.FCS_editor.save()
        if not self.SCS_editor.saved: self.SCS_editor.save()
        if not self.cv_net.saved: self.saveNet()
    
    def get_default_grid_path(self) -> str:
        return str(Path(self.folder) / Path(self.folder).name) + ".grid.xml"
    
    def get_default_roadnet_path(self) -> str:
        return str(Path(self.folder) / Path(self.folder).name) + ".net.xml"
    
    def saveNet(self):
        assert self.state is not None
        if self.state.grid:
            gpath = self.state.grid
            os.remove(gpath)
            if not gpath.lower().endswith(".xml"):
                gpath = self.get_default_grid_path()
        else:
            gpath = self.get_default_grid_path()
        if self.state.net:
            npath = self.state.net
            if not npath.lower().endswith(".xml"):
                npath = self.get_default_roadnet_path()
        else:
            npath = self.get_default_roadnet_path()
        self.cv_net.save(gpath, npath)
        
    def onDestroy(self):
        if not self.saved:
            ret = MB.askyesnocancel(_L["MB_INFO"], _L["MB_EXIT_SAVE"])
            if ret is None: return
            if ret: self.save()
        self.destroy()

    @errwrapper
    def simulate(self):
        if not self.__checkFolderOpened(): return
        start = try_int(self.entry_start.get(), "start time")
        assert start >= 0, "Start time must be non-negative integer"
        break_at = try_int(self.entry_break.get(), "break time")
        assert break_at > start, "Break time must be greater than start time"
        end = try_int(self.entry_end.get(), "end time")
        assert end >= break_at, "End time must be greater than or equal to break time"
        step = try_int(self.entry_step.get(), "time step")
        assert step > 0, "Time step must be positive integer"
        seed = try_int(self.entry_seed.get(), "random seed")
        assert self.state, "No project loaded"
        assert "scs" in self.state, "No SCS loaded"
        assert "fcs" in self.state, "No FCS loaded"
        assert "veh" in self.state, "No vehicles loaded"

        logs = []
        for x in ("fcs","scs","ev","gen","bus","line","pvw","ess"):
            if self.sim_statistic[x]:
                logs.append(x)
        assert logs, _L["NO_STA"]

        if not self.saved:
            if not MB.askyesno(_L["MB_INFO"],_L["MB_SAVE_AND_SIM"]): return
            self.save()
        
        # Save preference
        vcfg = V2SimConfig()
        vcfg.start_time = start
        vcfg.break_time = break_at
        vcfg.end_time = end
        vcfg.traffic_step = step
        vcfg.seed = seed
        vcfg.load_state = self.sim_load_state.get()
        vcfg.save_state_on_abort = self.sim_save_on_abort.get()
        vcfg.save_state_on_finish = self.sim_save_on_finish.get()
        vcfg.copy_state = self.sim_copy_state.get()
        vcfg.routing_method = self.ralgo.get()
        vcfg.disable_parallel = self.sim_no_parallel.get()
        vcfg.show_uxsim_info = self.sim_show_uxsim_info.get()
        vcfg.stats = logs
        vcfg.save(self.folder + "/preference.v2simcfg")
        
        if self.sim_load_state.get() == 0:
            cmd_load_state = ""
        elif self.sim_load_state.get() == 1:
            cmd_load_state = "--load-last-state"
        else:
            cmd_load_state = f'--initial-state="{self.folder}/saved_state"'
        commands = [sys.executable,
                    str(V2SIM_UX_DIR / "sim_single.py"),
                    f'-d="{self.folder}"', 
                    f"-b={start}", 
                    f"--break-at={self.entry_break.get()}",
                    f"-e={end}", 
                    f"-l={step}", 
                    "-log", ','.join(logs),
                    "--seed", str(seed),
                    cmd_load_state,
                    "--save-on-abort" if self.sim_save_on_abort.get() else "",
                    "--save-on-finish" if self.sim_save_on_finish.get() else "",
                    "--copy-state" if self.sim_copy_state.get() else "",
                    "--route-algo", self.ralgo.get(),
                    "--no-parallel" if self.sim_no_parallel.get() else "",
                    "--show-uxsim-info" if self.sim_show_uxsim_info.get() else "",
                    "--randomize-uxsim" if self.sim_randomize_traffic.get() else ""
                ]
        
        self.destroy()
        try:
            os.system(" ".join(commands))
        except KeyboardInterrupt:
            pass
        

    def savePlugins(self):
        def _save(data:List[tuple]):
            if not self.__checkFolderOpened():
                return False
            self.setStatus(_L["SAVE_PLG"])
            if self.state and "plg" in self.state:
                filename = self.state["plg"]
            else:
                filename = self.folder + "/plugins.plg.xml"
            try:
                rt = ET.Element("root")
                with open(filename, "w") as f:
                    f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                    for d in data:
                        attr = {"interval":str(d[1]), "enabled":str(d[2])}
                        attr.update(eval(d[4]))
                        for k,v in attr.items():
                            if not isinstance(v, str):
                                attr[k] = str(v)
                        e = ET.Element(d[0], attr)
                        if d[3] != ALWAYS_ONLINE:
                            ol = ET.Element("online")
                            lst = eval(d[3])
                            for r in lst:
                                ol.append(ET.Element("item", {"btime":str(r[0]), "etime":str(r[1])}))
                            e.append(ol)
                        rt.append(e)
                    f.write(ET.tostring(rt, "unicode", ).replace("><", ">\n<"))
                pass
            except Exception as e:
                self.setStatus(f"Error: {e}")
                traceback.print_exc()
                showerr(f"Error saving plugins: {e}")
                return False
            self.setStatus("Plugins saved")
            return True
        return _save
    
    def __checkFolderOpened(self):
        if not self.folder:
            showerr(_L["PROJ_NO_OPEN"])
            return False
        return True
    
    def _load_tg(self, after:OAfter=None):
        try:
            self.tg = TrafficGenerator(self.folder)
        except Exception as e:
            traceback.print_exc()
            showerr(f"Error loading traffic generator: {e}")
            self.tg = None
        else:
            if after: after()
    
    def _load(self, loads:Optional[List[str]] = None, async_:bool = True):
        if not self.folder:
            showerr("No project folder selected")
            return
        if loads is None: loads = [
            LOAD_GEN, LOAD_FCS, LOAD_SCS, LOAD_CSCSV, LOAD_NET, LOAD_PLG
        ]
        self._ldfrm = LoadingBox(loads, self._Q)
        self.update()
        self.after(100, self.__load_part2, set(loads), async_)
    
    def __load_part2(self, loads:Set[str], async_:bool):
        self.state = res = DetectFiles(self.folder)
        self.title(f"{_L['TITLE']} - {Path(self.folder).name}")
        self.update()

        # Check if grid exists
        if not res.grid: 
            with open(self.get_default_grid_path(),"w") as f:
                f.write(DEFAULT_GRID)
            self.state = res = DetectFiles(self.folder)
        
        self.update()
        
        # Load traffic generator
        if LOAD_GEN in loads:
            self._Q.submit("TrafficGenLoaded", self._load_tg)

        self.update()

        # Load FCS
        if LOAD_FCS in loads:
            self._load_fcs(lambda: self._ldfrm.setText(LOAD_FCS, _L['DONE']))

        self.update()

        # Load SCS
        if LOAD_SCS in loads:
            self._load_scs(lambda: self._ldfrm.setText(LOAD_SCS, _L['DONE']))
        
        self.update()

        # Load CSCSV
        if LOAD_CSCSV in loads:
            self._load_cscsv(lambda: self._ldfrm.setText(LOAD_CSCSV, _L['DONE']))
        
        self.update()

        # Load plugins
        if LOAD_PLG in loads:
            self._load_plugins()
            self._ldfrm.setText(LOAD_PLG,_L['DONE'])
        
        self.update()

        self.rb_veh_src2.configure(state="normal" if "poly" in res else "disabled")
        self.rb_veh_src1.configure(state="normal" if "taz" in res else "disabled")
        
        self.state = res = DetectFiles(self.folder)

        self.update()
        
        if LOAD_NET in loads:
            self.cv_net.clear()
            self._load_network(self.tabs.select(),
                lambda: self._ldfrm.setText(LOAD_NET, _L['DONE']))
        
        self.update()

        def setText(lb:Label, itm:str, must:bool = False):
            if itm in res:
                lb.config(text=Path(res[itm]).name, foreground="black")
            else:
                lb.config(text="None", foreground="red" if must else "black")
        
        setText(self.lb_fcs, "fcs", True)
        setText(self.lb_scs, "scs", True)
        setText(self.lb_grid, "grid")
        setText(self.lb_net, "net", True)
        setText(self.lb_veh, "veh", True)
        setText(self.lb_plg, "plg")
        setText(self.lb_py, "py")
        setText(self.lb_node_type, "node_type")
        setText(self.lb_osm, "osm")
        setText(self.lb_poly, "poly")
        setText(self.lb_poi, "poi")
        setText(self.lb_cscsv, "cscsv")

        self.update()
        
        if self.state.pref:
            vcfg = V2SimConfig.load(self.state.pref)
            self.entry_start.delete(0, END)
            self.entry_start.insert(0, str(vcfg.start_time))
            self.entry_break.delete(0, END)
            self.entry_break.insert(0, str(vcfg.break_time))
            self.entry_end.delete(0, END)
            self.entry_end.insert(0, str(vcfg.end_time))
            self.entry_step.delete(0, END)
            self.entry_step.insert(0, str(vcfg.traffic_step))
            self.entry_seed.delete(0, END)
            self.entry_seed.insert(0, str(vcfg.seed))
            self.ralgo.set(vcfg.routing_method)
            self.sim_load_state.set(vcfg.load_state)
            self.on_load_state_changed()
            self.sim_save_on_finish.set(vcfg.save_state_on_finish)
            self.sim_save_on_abort.set(vcfg.save_state_on_abort)
            self.sim_copy_state.set(vcfg.copy_state)
            self.sim_no_parallel.set(vcfg.disable_parallel)
            self.sim_show_uxsim_info.set(vcfg.show_uxsim_info)
            if vcfg.stats:
                for x in vcfg.stats:
                    if x in self.sim_statistic:
                        self.sim_statistic[x] = True
                    else:
                        showerr(_L["UKN_STA_TYPE"].format(x, ', '.join(self.sim_statistic.keys())))
        
        self.sim_cb_load_last_state.configure(state="normal" if self.state.last_result_state else "disabled")
        self.sim_cb_load_saved_state.configure(state="normal" if self.state.saved_state else "disabled")

        self.update()

        self.setStatus(_L["STA_READY"])
        
        if len(loads) == 0: self._ldfrm.destroy()
    
    def _load_plugins(self):
        plg_set:Set[str] = set()
        plg_enabled_set:Set[str] = set()

        self.sim_plglist.clear()
        assert self.state is not None
        if self.state.plg:
            et = ReadXML(self.state.plg)
            if et is None:
                showerr(_L["ERR_LOAD_PLG"])
                return
            rt = et.getroot()
            if rt is None:
                showerr(_L["ERR_LOAD_PLG"])
                return
            for p in rt:
                try:
                    plg_type = self.sim_plglist.plg_pool.GetPluginType(p.tag.lower())
                except KeyError:
                    plg_list = ', '.join(self.sim_plglist.plg_pool.GetAllPlugins().keys())
                    showerr(_L["UKN_PLG_TYPE"].format(p.tag, plg_list))
                    continue
                assert issubclass(plg_type, PluginBase), "Plugin type is not a subclass of PluginBase"

                attr = plg_type.ElemShouldHave().default_value_dict()
                plg_set.add(p.tag.lower())

                # Check online attribute
                olelem = p.find("online")
                if olelem is not None: ol_str = RangeList(olelem)
                else: ol_str = ALWAYS_ONLINE

                # Check enabled attribute
                enabled = p.attrib.pop("enabled", SIM_YES)
                if enabled.upper() != SIM_NO:
                    enabled = SIM_YES
                    plg_enabled_set.add(p.tag.lower())
                
                # Check interval attribute
                intv = p.attrib.pop("interval")
                attr.update(p.attrib)
                self.sim_plglist.add(p.tag, intv, enabled, ol_str, attr)

        # Check if PDN exists
        if "pdn" not in plg_set:
            pdn_attr_default = PluginPDN.ElemShouldHave().default_value_dict()
            self.sim_plglist.add("pdn", 300, SIM_YES, ALWAYS_ONLINE, pdn_attr_default)
            plg_set.add("pdn")
            plg_enabled_set.add("pdn")
        
        # Check if V2G exists
        if "v2g" not in plg_set:
            self.sim_plglist.add("v2g", 300, SIM_YES, ALWAYS_ONLINE, {})
        if not self.state.plg:
            self.sim_plglist.save()
        
        self.__OnPluginEnabledSet()
        
    def _load_fcs(self, afterx:OAfter = None):
        assert self.state is not None
        def after():
            assert self.state is not None
            v = "fcs" in self.state
            self.sim_statistic["fcs"]=v
            self.sim_statistic.setEnabled("fcs", v)
            self.FCS_editor.setPoly("poly" in self.state)
            self.FCS_editor.setCSCSV("cscsv" in self.state)
            if afterx: afterx()
        if self.state.fcs:
            self.FCS_editor._Q.setcallback("loaded", after)
            self.FCS_editor.load(self.state.fcs)
        else:
            self.FCS_editor.clear()
            after()
        
    def _load_scs(self, afterx:OAfter=None):
        assert self.state is not None
        def after():
            assert self.state is not None
            v = "scs" in self.state
            self.sim_statistic["scs"] = v
            self.sim_statistic.setEnabled("scs", v)
            self.SCS_editor.setPoly("poly" in self.state)
            self.SCS_editor.setCSCSV("cscsv" in self.state)
            if afterx: afterx()
        if self.state.scs:
            self.SCS_editor._Q.setcallback("loaded", after)
            self.SCS_editor.load(self.state.scs)
        else:
            self.SCS_editor.clear()
            after()

    def _load_cscsv(self, after:OAfter = None):
        assert self.state is not None
        if self.state.cscsv:
            if after:
                self.CsCsv_editor._Q.setcallback("loaded", after)
            self.CsCsv_editor.load(self.state.cscsv)
        else:
            self.CsCsv_editor.clear()
            if after: after()
    
    def _load_network(self, tab_ret, after:OAfter = None):
        if self.state and self.state.net:
            self.tabs.select(self.tab_Net)
            time.sleep(0.01)
            self.tabs.select(tab_ret)
            self.lb_gridsave.config(text=_L["SAVED"],foreground="green")
            
            assert self.state is not None and self.state.net is not None
            if self.state.grid:
                self.cv_net.setGrid(PowerGrid.fromFile(self.state.grid))
            assert self.cv_net.Grid is not None
            self.lb_puvalues.configure(text=_L["PU_VALS"].format(self.cv_net.Grid.Ub,self.cv_net.Grid.Sb_MVA))

            def __el(state: FileDetectResult, after:OAfter=None):
                assert state.net is not None
                el = RoadNet.load(state.net)
                return el, after
            
            self._Q.submit("cvnetloaded", __el, self.state, after)

    def on_cvnet_loaded(self, el:RoadNet, after:OAfter=None):
        self.cv_net.setRoadNet(el, after = after)

    def openFolder(self):
        init_dir = Path("./cases")
        if not init_dir.exists(): init_dir.mkdir(parents=True, exist_ok=True)
        folder = filedialog.askdirectory(initialdir=str(init_dir),mustexist=True,title="Select project folder")
        if folder:
            self.folder = str(Path(folder))
            self._load()
    
    def generateCS(self, ctl:CSEditorGUI, cscsv_mode:int, **kwargs):
        if not self.tg:
            showerr("No traffic generator loaded")
            return
        self.setStatus("Generating CS...")
        if cscsv_mode == 0:
            cs_file = ""
            poly_file = ""
        elif cscsv_mode == 1:
            cs_file = self.state.cscsv if self.state else ""
            poly_file = ""
        else:
            cs_file = ""
            poly_file = self.state.poly if self.state else ""
        use_grid = kwargs.pop("use_grid", False)
        assert self.state is not None
        if use_grid:
            if self.state.grid is None:
                showerr("No grid loaded")
                return
            kwargs["grid_file"] = self.state.grid
        kwargs["cs_file"] = cs_file
        kwargs["poly_file"] = poly_file

        def work(ctl, **kwargs):
            try:
                if not self.tg: return
                warns, far_cnt, scc_cnt = self.tg._CS(**kwargs)
                if kwargs["mode"] == "fcs":
                    self._load([LOAD_FCS, LOAD_GEN])
                else:
                    self._load([LOAD_SCS, LOAD_GEN])
                return ctl, None, warns, far_cnt, scc_cnt
            except Exception as e:
                print(f"\nError generating CS: {e}")
                traceback.print_exc()
                return ctl, e, [], 0, 0
            
        self._Q.submit("CSGenDone", work, ctl, **kwargs)   
    
    @errwrapper
    def generateVeh(self):
        assert self.tg, _L["MSG_NO_TRAFFIC_GEN"]
        if not self.__checkFolderOpened(): return
        self.setStatus(_L["STA_GEN_VEH"])
        carcnt = try_int(self.entry_carcnt.get(), "Vehicle count")
        carseed = try_int(self.entry_carseed.get(), "Vehicle seed")
        day_count = try_int(self.entry_daycnt.get(), "Vehicle day count")
        
        try:
            pars = self.veh_pars.getAllData()
            new_pars = {
                "v2g_prop":float(self.entry_v2gprop.get()),
                "omega":eval(pars["Omega"]),
                "krel":eval(pars["KRel"]),
                "ksc":eval(pars["KSC"]),
                "kfc":eval(pars["KFC"]),
                "kv2g":eval(pars["KV2G"]),
            }
        except Exception as e:
            raise ValueError("Invalid Vehicle parameters") from e
        
        if self.veh_gen_src.get() == 0:
            mode = TripsGenMode.AUTO
        elif self.veh_gen_src.get() == 1:
            mode = TripsGenMode.TYPE
        else:
            mode = TripsGenMode.POLY
        route_cache = RoutingCacheMode(self.veh_route_cache.get())
        self.btn_genveh.config(state = DISABLED)

        def work() -> Optional[Exception]:
            try:
                assert self.tg
                self.tg.EVTrips(carcnt, carseed, day_count, mode = mode, route_cache = route_cache, **new_pars)
                self._load([])
                return None
            except Exception as e:
                return e
        
        self._Q.submit("VehGenDone", work)

    def draw(self):
        if not self.__checkFolderOpened(): return
        self.cv_net.UnlocateAllEdges()
        s = set(x.strip() for x in self.entry_Ledges.get().split(','))
        self.cv_net.LocateEdges(s)

    def CSCSVDownloadWorker(self):
        if not self.__checkFolderOpened(): return
        self.setStatus("Downloading CS CSV...")
        key = self.CsCsv_editor.entry_amapkey.get()
        def work():
            try:
                csQuery(self.folder,"",key,True)
                self._load([LOAD_CSCSV])
                return None
            except Exception as e:
                return e
        
        self._Q.submit("CSCSVDownloadDone", work)
        
    def setStatus(self, text:str):
        self.sbar.config(text=text)

    def _win(self):
        self.title(_L["TITLE"])

__all__ = ["MainBox"]